package Packer;

require 'agent.include.pl';
# makePasswordNode, generateDumpId, XmlNode, makeXmlNode,
# makeMIMEBase64, urlDecode,
# getSqlList

use strict;
use warnings;

use Status;
use HelpFuncs;
use Storage::FileNameCreator;

use vars qw|@ISA|;

require "agent.include.pl";


my $DEBUG = undef;

my %clientsAttribute = (
  'name'    => 'login',
  'contact' => 'pname',
  'cr-date' => 'cr_date',
);

my %clientsInfo = (
  'company' => 'cname',
  'phone'   => 'phone',
  'fax'     => 'fax',
  'address' => 'address',
  'city'    => 'city',
  'state'   => 'state',
  'zip'     => 'pcode',
  'country' => 'country',
  'email'   => 'email',
);

my %adminInfo = (
  'cname' => 'company',
  'phone'   => 'phone',
  'fax'     => 'fax',
  'address' => 'address',
  'city'    => 'city',
  'state'   => 'state',
  'pcode'   => 'zip',
  'country' => 'country',
  'email'   => 'email',
  'pname'   => 'name',
  'locale'  => 'locale',
);

my %locale_map = (
  'bg' => 'bg',
  'hk' => 'zh-HK',
  'ru' => 'ru-RU',
  'en' => 'en-US',
  'nl' => 'nl-NL',
  'br' => 'pt-BR',
  'it' => 'it-IT',
  'tr' => 'tr-TR',
  'es' => 'es-ES',
  'pl' => 'pl-PL',
  'ca' => 'ca-ES',
  'jp' => 'ja-JP',
  'tw' => 'zh-TW',
  'fi' => 'fi-FI',
  'cn' => 'zh-CN',
  'ko' => 'ko-KR',
  'fr' => 'fr-FR',
  'pt' => 'pt-PT',
  'de' => 'de-DE',
  'lt' => 'lt-LT'
);
#list of attributes for template-item node in dump . xml

  # needs to synchronize with template-item's ATTLIST of plesk.dtd
  my %template_items = (
  'webmail'         => 1,
  'disk_space'      => 1,
  'wu_script'       => 1,
  'stat_ttl'        => 1,
  'maillists'       => 1,
  'nonexist_mail'   => 1,
  'catch_addr'      => 1,
  'bounce_mess'     => 1,
  'max_traffic'     => 1,
  'max_box'         => 1,
  'mbox_quota'      => 1,
  'max_redir'       => 1,
  'max_mg'          => 1,
  'max_resp'        => 1,
  'max_wu'          => 1,
  'max_db'          => 1,
  'max_maillists'   => 1,
  'max_webapps'     => 1,
  'max_subdom'      => 1,
  'expiration'      => 1,
  'vh_type'         => 1,
  'quota'           => 1,
  'ssl'             => 1,
  'same_ssl'        => 1,
  'fp'              => 1,
  'fp_ssl'          => 1,
  'fp_auth'         => 1,
  'asp'             => 1,
  'ssi'             => 1,
  'php'             => 1,
  'cgi'             => 1,
  'perl'            => 1,
  'python'          => 1,
  'coldfusion'      => 1,
  'webstat'         => 1,
  'errdocs'         => 1,
  'shell'           => 1,
  'pdir_plesk_stat' => 1,
  'dns_type'        => 1,
  'max_dom_aliases' => 1,

  # client template
  'allow_ftp_backups'       => 1,
  'allow_local_backups'     => 1,
  'change_limits'           => 1,
  'cp_access'               => 1,
  'create_domains'          => 1,
  'dashboard'               => 1,
  'excl_ip_num'             => 1,
  'manage_anonftp'          => 1,
  'manage_crontab'          => 1,
  'manage_dashboard'        => 1,
  'manage_dns'              => 1,
  'manage_domain_aliases'   => 1,
  'manage_drweb'            => 1,
  'manage_log'              => 1,
  'manage_maillists'        => 1,
  'manage_not_chroot_shell' => 1,
  'manage_phosting'         => 1,
  'manage_quota'            => 1,
  'manage_sh_access'        => 1,
  'manage_spamfilter'       => 1,
  'manage_subdomains'       => 1,
  'manage_webapps'          => 1,
  'max_dom'                 => 1,
  'remote_access_interface' => 1,
  'stdgui'                  => 1,
  'manage_php_safe_mode'    => 1,
  'manage_webstat'          => 1
  );

my %domainUserAttribute = ( 'contact' => 'personalName', );

my %domainUserInfo = (
  'company' => 'companyName',
  'phone'   => 'phone',
  'fax'     => 'fax',
  'address' => 'address',
  'city'    => 'city',
  'state'   => 'state',
  'zip'     => 'zip',
  'email'   => 'email',
  'country' => 'country',
);

my %hostingAttribute = (
  'https'   => 'ssl',
  'fp'      => 'fp',
  'fpssl'   => 'fp_ssl',
  'fpauth'  => 'fp_enable',
  'webstat' => 'webstat',

  #                        'at-domains'=>'at_domains',
);



my %emptyableBoolElementInTemplate = (
  'asp'             => 1,
  'cgi'             => 1,
  'coldfusion'      => 1,
  'pdir_plesk_stat' => 1,
  'perl'            => 1,
  'php'             => 1,
  'python'          => 1,
  'ssi'             => 1
);

my %webUserScripting = (
  'ssi'         => 'ssi',
  'php'         => 'php',
  'cgi'         => 'cgi',
  'perl'        => 'perl',
  'asp'         => 'asp',
  'python'      => 'python',
  'asp_dot_net' => 'asp_dot_net',
  'fastcgi'     => 'fastcgi',
  'miva'        => 'miva'
);

### Common function ###

sub new {
  my $self = {};
  bless( $self, shift );
  $self->_init(@_);
  return $self;
}

sub _init {
  my ( $self, $version, $storagePolicy ) = @_;

  XmlNode::resetCompatibilityProcs();

  $self->{storage} = $storagePolicy;
  $self->{version} = $version;
  $self->{base64} = makeMIMEBase64();
  $self->{fnamecreator} = Storage::FileNameCreator->new();
  $self->setBackupProfileFileName( 'backup' );
  $self->{migration} = 1;
  $self->{ownerGuid} = '';
  $self->{ownerType} = '';
  $self->{roots} = [];
}

sub setBackupProfileFileName{
  my ($self, $profileName, $profileId ) = @_;
  $self->{backupname} = $self->{fnamecreator}->normalize_long_string( $profileName, $profileId );
  $profileId = '' if not defined $profileId;
  printToLog( "Set backup file name '$self->{backupname}' (profile '$profileName', id='$profileId')\n" );
}

sub setBackupOwnerGuid{
  my ($self, $ownerGuid, $ownertype ) = @_;
  $ownerGuid = '' if not $ownerGuid;
  $self->{ownerGuid} = $ownerGuid;
  $ownertype = '' if not $ownertype;
  $self->{ownerType} = $ownertype;
  printToLog( "Set backup owner guid '$ownerGuid', type '$ownertype'\n" );
}

sub getBackupOwnerGuid{
  my ($self) = @_;
  return $self->{ownerGuid};
}

sub startCollectStorageStatistics {
  my $self = shift;

  $self->{storage}->startCollectStatistics();
}

sub getStorageStatistics {
  my $self = shift;

  return $self->{storage}->getStatistics();
}

sub setRoot {
  my ( $self, $description, $content ) = @_;

  $self->{root} = XmlNode->new( 'migration-dump',
    'attributes' =>
      { 'agent-name' => 'PleskX', 'dump-version' => $self->{version} } );
 $self->{root}->setAttribute( 'content-included', $content ? 'true' : 'false' ) if defined $content;
  my $dumpinfo = $self->{root}->getChild( 'dump-info', 1 );
  if ($description) {
    $dumpinfo->addChild( XmlNode->new( 'description', 'content' => $description ) );
  }
}

sub turnOffMigrationMode{
  my ($self) = @_;
  $self->{migration} = 0;
  printToLog( "The migration mode is turn off" );
}

sub tunOffContent{
  my ($self) = @_;
  $self->{skip_content} = 1;
  printToLog( "The dump of content is switched off" );
}

sub isRootNode{
  my ($self, $node) = @_;
  foreach my $root( @{$self->{roots}} ){
    return 1 if $root==$node;
  }
  return 0;
}

sub finishChild{
    my ($self, $node, $path, $fileName ) = @_;
    my $ret;
    if( scalar( @{$self->{roots}} )==1  && $self->isRootNode($node) ) {
    	$ret = $self->{storage}->finish( $self->{root}, $path, $fileName );
        die "Cannot create main dump file!" if $ret!=0;
        $ret = $self->{storage}->getMainDumpXmlFile();
     }
    else {
    	$ret = $self->{storage}->finishChild( $self->{root}, $node, $path, $fileName );
    }
    return $ret;
}

sub finish {
    my ($self) = @_;

    my %infoxml;
    my $fileid;
    my $ret = 0;
    my $fileName;

    if( $self->{migration} || scalar( @{$self->{roots}} )!=1 ){
        XmlNode::setStartSavePath( '' );
        if( $self->{migration} ) { $fileName = 'dump'; }
        else {  $fileName = $self->{fnamecreator}->getFileName( '', $self->{backupname}, '', 'info' ); }
        $ret = $self->{storage}->finish( $self->{root}, '', $fileName );
        die "Cannot create main dump file!" if $ret!=0;
        return $ret if $self->{migration};
        @{$infoxml{$self->{storage}->getMainDumpXmlFile()}} = ( $self->{root}->getContentSize(), $self->getAdminObjectId(), ( exists $self->{admin} ? $self->{admin}->getGuid() : $self->{ownerGuid} ) );
        $self->{storage}->createRepositoryIndex( ( exists $self->{admin} ? $self->{admin}->getGuid() : '' ) . '_' . $self->getAdminObjectId() );
    }
    $fileName = 'info';

    my ( $id, $node, $path );
    if( exists $self->{clientNodes} ){
      while( ($id, $node ) = each( %{$self->{clientNodes}} ) ){
         my $fname;
         if( exists $self->{resellersNodes}->{$id} ) {
           $path = $self->getResellersBackupPath( $id );
           $fname = $self->getResellersBackupPath( $id, $fileName, 1 );
           printToLog( "Save reseller dump $path" );
         }
         else {
           $path = $self->getClientsBackupPath( $id );
           $fname = $self->getClientsBackupPath( $id, $fileName, 1 );
           printToLog( "Save client dump $path" );
         }
         XmlNode::setStartSavePath( $path );
         $fileid = $self->finishChild( $node, $path, $fname );
         @{$infoxml{$fileid}} = ( $node->getContentSize(), $self->getClientObjectId( $id ), $node->getGuid() );
         $self->{storage}->createRepositoryIndex( $node->getGuid() . '_' . $self->getClientObjectId( $id ) );
      }
    }
    if( exists $self->{domainNodes} ) {
      while( ($id, $node ) = each( %{$self->{domainNodes}} ) ){
         $path = $self->getDomainsBackupPath( $id );
         printToLog( "Save domain dump $path" );
         XmlNode::setStartSavePath( $path );
         $fileid = $self->finishChild( $node, $path, $self->getDomainsBackupPath( $id, $fileName, 1 ) );
         @{$infoxml{$fileid}} = ( $node->getContentSize(), $self->getDomainObjectId( $id ), $node->getGuid() );
         $self->{storage}->createRepositoryIndex( $node->getGuid() . '_' . $self->getDomainObjectId( $id ) );
      }
    }
#    if( exists $self->{phostingNodes} ){
#      while( ($id, $node ) = each( %{$self->{phostingNodes}} ) ){
#         $path = $self->getPhostingBackupPath( $id );
#         printToLog( "Save phosting dump $path" );
#         XmlNode::setStartSavePath( $path );
#         $fileid = $self->finishChild( $node, $path, $self->getPhostingBackupPath( $id, $fileName, 1 ) );
#         @{$infoxml{$fileid}} = ( $node->getContentSize(), $self->getPhostingObjectId( $id ), $node->getGuid() );
#         $self->{storage}->createRepositoryIndex( $node->getGuid() . '_' . $self->getPhostingObjectId( $id ) );
#      }
#    }
#    if( exists $self->{subDomainNodes} ){
#      while( ($id, $node ) = each( %{$self->{subDomainNodes}} ) ) {
#         $path = $self->getSubdomainsBackupPath( $id );
#         printToLog( "Save subdomain dump $path" );
#         XmlNode::setStartSavePath( $path );
#         $fileid = $self->finishChild( $node, $path, $self->getSubdomainsBackupPath( $id, $fileName, 1 ) );
#         @{$infoxml{$fileid}} = ( $node->getContentSize(), $self->getSubdomainObjectId( $id ), $node->getGuid() );
#         $self->{storage}->createRepositoryIndex( $node->getGuid() . '_' . $self->getSubdomainObjectId( $id ) );
#      }
#    }
#    if( exists( $self->{mailnamesNodes} ) ) {
#      while( ($id, $node ) = each( %{$self->{mailnamesNodes}} ) ){
#         $path = $self->getMailnamesBackupPath( $id );
#         printToLog( "Save mail name dump $path" );
#         XmlNode::setStartSavePath( $path );
#         $fileid = $self->finishChild( $node, $path, $self->getMailnamesBackupPath( $id, $fileName, 1 ) );
#         @{$infoxml{$fileid}} = ( $node->getContentSize(), $self->getMailnameObjectId( $id ), $node->getGuid() );
#         $self->{storage}->createRepositoryIndex( $node->getGuid() . '_' . $self->getMailnameObjectId( $id ) );
#      }
#    }
#    if( exists $self->{databaseNodes} ){
#      while( ($id, $node ) = each( %{$self->{databaseNodes}} ) ){
#         $path = $self->getDatabasesBackupPath( $id );
#         printToLog( "Save database dump $path" );
#         XmlNode::setStartSavePath( $path );
#         $fileid = $self->finishChild( $node, $path, $self->getDatabasesBackupPath( $id, $fileName, 1 ) );
#         @{$infoxml{$fileid}} = ( $node->getContentSize(), $self->getDatabaseObjectId( $id ), $node->getGuid() );
#         $self->{storage}->createRepositoryIndex( $node->getGuid() . '_' . $self->getDatabaseObjectId( $id ) );
#      }
#    }

    #Create discovered files
    foreach $fileid(  keys( %infoxml ) ){
       my $idx = rindex( $fileid, '/' );
       $path = $fileid;
       $path = substr( $path, 0, $idx ) if $idx>0;
       my $out = '';
       my $xmlsize = 0;
       my $files;
       foreach $id( keys( %infoxml ) ){
         if( $idx<0 || index( $id, $path )==0 ){
            $files = $self->{storage}->getFilesFromId( $id );
            foreach my $filedata( @{$files} ) {
              $xmlsize += $filedata->[1];
            }
         }
       }
       $files = $self->{storage}->getFilesFromId( $fileid );
       if( scalar( @{$files} )==1 ){
          my $xmlfiledata = $files->[0];
          $xmlsize += @{$infoxml{$fileid}}[0];
          my $objId = @{$infoxml{$fileid}}[1];
          my $objGuid = @{$infoxml{$fileid}}[2];
          $self->{storage}->writeDiscovered( $self->{storage}->getFilePathFromId( $fileid ), $xmlfiledata->[0], $xmlsize, $self->{ownerGuid}, $self->{ownerType}, $objGuid, $objId );
       }
    }

    return 0;
}


### Server functions ###

sub setServerSettings {
  my ( $self ) = @_;

  my $root = $self->{root};
  unless (defined $root)
  {
    printToError("Root node is not set");
    return;
  }

  my $serverNode = XmlNode->new('server');

  $self->{serverNode} = $serverNode;

  $root->addChild($serverNode);
}

sub addServerCustomButton {
  my ( $self, $id, $optionsPtr, $customButtonsDir, $icon ) = @_;

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: addServerCustomButton: empty parent node');
    return;
  }

  my $node = $self->makeCustomButtonNode( 'admin', undef, $id, $optionsPtr, $customButtonsDir, $icon );

  $parent->getChild( 'interface-preferences', 1 )->addChild($node);
}

sub addServerIp {
  my ( $self, $ip ) = @_;

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: addServerIp: empty parent node');
    return;
  }

  $self->makeSystemIpNode( $parent, $ip );
}

sub setServerDefaultIp {
  my ( $self, $ip ) = @_;

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: addServerDefaultIp: empty parent node');
    return;
  }

  $parent->getChild( 'properties', 1 )->addChild( XmlNode->new( 'default-ip', 'content' => $ip ) );
}

sub setServerHostname {
  my ( $self, $hostname ) = @_;

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: setServerHostname: empty parent node');
    return;
  }

  $parent->getChild( 'properties', 1 )->addChild( XmlNode->new( 'hostname', 'content' => $hostname ) );
}

sub setServerAdminInfo {
  my ( $self, $ptrAdmin, $passwd, $skin, $max_btn_len, $send_announce ) = @_;

  my $parent = $self->{admin};
  if ( !defined($parent) ) {
    printToError('Error: setServerAdminInfo: empty parent node');
    return;
  }

  $parent->setAttribute( 'skin', $skin ) if $skin;
  $parent->setAttribute( 'max-button-length', $max_btn_len ) if $max_btn_len;
  $parent->setAttribute( 'send-announce', $send_announce ) if $send_announce;

  my $pref = $parent->getChild( 'preferences', 1 );
  my ( $name, $value );
  $pref->addChild(  XmlNode->new( 'admin-password', 'content' => $passwd ) );
  while ( ( $name, $value ) = each( %{$ptrAdmin} ) ) {

    #Attributes transformation. Bug 97408
    $name =~ s/^admin_//g;
    next if not defined $adminInfo{$name};
    $pref->addChild( XmlNode->new( 'pinfo', 'attributes' => { 'name' => $adminInfo{$name} }, 'content'    => $value ) );
  }
}

sub adddServerDb {
  my ( $self, $dbServerPtr, $passwd, $default ) = @_;

  my %dbServer = %{$dbServerPtr};

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: adddServerDb: empty parent node');
    return;
  }

  my $dbServerNode = XmlNode->new('db-server');
  my $param        = 'default_server_' . $dbServer{'type'};

  if ( $default ) {
    $dbServerNode->setAttribute( 'default', 'true' );
  }
  $dbServerNode->setAttribute( 'type', $dbServer{'type'} );

  $dbServerNode->addChild( XmlNode->new( 'host', 'content' => "$dbServer{'host'}" ) );
  $dbServerNode->addChild( XmlNode->new( 'port', 'content' => "$dbServer{'port'}" ) );

  if ( ($dbServer{'admin_login'}) and ($dbServer{'admin_login'} ne '')) {
    my $adminNode =
      XmlNode->new( 'db-admin',
      "attributes" => { "name" => "$dbServer{'admin_login'}" } );
    if ( $dbServer{'type'} eq 'mysql'
      && $dbServer{'host'} eq 'localhost' )
    {
      $dbServer{'admin_password'} = $passwd;
    }
    my $passwordNode =
      makePasswordNode( $dbServer{'admin_password'}, 'plain' )
      ;    # password type for dbservers is always plain
    $adminNode->addChild($passwordNode);

    $dbServerNode->addChild($adminNode);
  }

  $parent->getChild( 'db-servers', 1 )->addChild($dbServerNode);
}

sub addServerKey {
  my ( $self, $keyId, $keyName, $keyDir, $additional ) = @_;

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: addServerKey: empty parent node');
    return;
  }

  if ( -e $keyDir . "/" . $keyName ) {

    my $keyNode =
      XmlNode->new( 'key',
      'attributes' => { 'additional' => $additional } );

    my $cid = $self->addAdminTar(
      'key',
      undef,
      "$keyId.key",
      "directory" => $keyDir,
      "include"   => [$keyName]
    );
    $keyNode->getChild( 'content', 1, 1 )->addChild($cid) if $cid;

    $parent->getChild( 'keys', 1 )->addChild($keyNode);
  }

}

sub setServerMail {
  my ( $self, $letterSize, $paramsPtr, $blackListPtr, $whiteListPtr ) = @_;

  my %params = %{$paramsPtr};
  my @blackList = @{$blackListPtr};
  my @whiteList = @{$whiteListPtr};

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: setServerMail: empty parent node');
    return;
  }

  my $mailNode = XmlNode->new(
    'mail-settings',
    'attributes' => {
      'max-letter-size' => $letterSize,
      'relay'           => $params{'relay'},
      'use-vocabulary'  => (defined($params{'use_vocabulary'}) and ($params{'use_vocabulary'} eq 'true'))
        ? 'true'
        : 'false',
      'short-pop3-names' => (defined($params{'allow_short_pop3_names'}) and ($params{'allow_short_pop3_names'})) eq 'enabled'
        ? 'true'
        : 'false',
      'message-submission' => (defined($params{'message_submission'}) and ($params{'message_submission'} eq 'true'))
        ? 'true'
        : 'false',
      'sign-outgoing-mail' => (defined($params{'domain_keys_sign'}) and ($params{'domain_keys_sign'} eq 'true'))
        ? 'true'
        : 'false',
      'verify-incoming-mail' => (defined($params{'domain_keys_verify'}) and ($params{'domain_keys_verify'} eq 'true'))
        ? 'true'
        : 'false',
    }
  );

  my $spfNode = XmlNode->new(
    'spf',
    'attributes' => {
      'status' => (
        exists $params{'spf_enabled'} and $params{'spf_enabled'} eq 'true'
        ) ? 'true' : 'false'
    }
  );
  $spfNode->setAttribute( 'spf-behavior',
    ( "$params{'spf_behavior'}" ne '' ) ? $params{'spf_behavior'} : 1 );

  $spfNode->addChild( XmlNode->new( 'spf-rules', 'content' => $params{'spf_rules'} ) );
  $spfNode->addChild( XmlNode->new( 'spf-guess', 'content' => $params{'spf_guess'} ) );
  $spfNode->addChild( XmlNode->new( 'spf-exp', 'content' => $params{'spf_exp'} ) );

  $mailNode->addChild($spfNode);

  my $rblNode = XmlNode->new(
    'rbl',
    'attributes' => {
        'status' => ( exists $params{'rbl'} and $params{'rbl'} eq 'true' )
      ? 'true'
      : 'false'
    }
  );
  $rblNode->addChild( XmlNode->new( 'rbl-server', 'content' => $params{'rbl_server'} ) );

  $mailNode->addChild($rblNode);

  if ( $params{'relay'} eq 'auth' ) {
    $mailNode->setAttribute( 'pop-auth', $params{'disable_pop_auth'} ? 'false' : 'true' );
    $mailNode->setAttribute( 'smtp-auth', $params{'disable_smtp_auth'} ? 'false' : 'true' );
    $mailNode->setAttribute( 'poplock-time', $params{'poplock_time'} )  if ( !$params{'disable_pop_auth'} );
  }

  # black list
  my $listNode = XmlNode->new('black-list');

  for my $ptrRow (@blackList) {
    $listNode->addChild( XmlNode->new( 'list-item', 'content' => $ptrRow->[0] ) );
  }
  $mailNode->addChild($listNode);

  # white list
  $listNode = XmlNode->new('white-list');

  for my $ptrRow (@whiteList) {
    $listNode->addChild(
      XmlNode->new(
        'list-item', 'content' => $ptrRow->[0] . '/' . $ptrRow->[1]
      )
    );
  }
  $mailNode->addChild($listNode);

  $parent->addChild($mailNode);
}

sub setServerDNS {
  my ( $self, $paramsPtr, $recordsPtr ) = @_;

  my @records = @{$recordsPtr};

  my %params = %{$paramsPtr};

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: setServerDNS: empty parent node');
    return;
  }

  # dump dns
  my $dnsNode = XmlNode->new(
    'dns-settings',
    'attributes' => {
      'recursion' => defined $params{'dns_recursion'}
      ? $params{'dns_recursion'}
      : 'any'
    }
  );

  my $dnsZone = XmlNode->new(
    'dns-zone',
    'attributes' =>
      { 'email' => 'root@localhost.localdomain', 'type' => 'master' },
    'children' =>
      [ Status::make( $params{'dns_zone_status'} ne 'false' ? 0 : 16 ) ]
  );

  $dnsZone->setAttribute( 'serial-format', $params{'soa_serial_format'} ) if exists $params{'soa_serial_format'};

  my %zone_params = (
    'ttl'     => 1 * 86400,
    'refresh' => 3 * 3600,
    'retry'   => 1 * 3600,
    'expire'  => 7 * 86400,
    'minimum' => 3 * 3600
  );

  my %zone_units = (
    'ttl'     => 86400,
    'refresh' => 3600,
    'retry'   => 3600,
    'expire'  => 86400,
    'minimum' => 3600
  );

  foreach my $param ( keys %zone_params ) {
    my $soa_param = 'soa_' . $param;

    $dnsZone->addChild(
      $self->makeDnsZoneParam(
        $param,
        exists $params{ $soa_param . '_unit' }
        ? $params{ $soa_param . '_unit' }
        : $zone_units{$param},
        exists $params{$soa_param} ? $params{$soa_param}
        : $zone_params{$param},
      )
    );
  }

  # dns records
  for my $ptrHash (@records) {
    my $dnsrec = $self->makeDnsRecord($ptrHash);
    if ($dnsrec) {
      $dnsZone->addChild($dnsrec);
    }
  }

  $dnsNode->addChild($dnsZone);

  # dump common acl for dns zone

  my $aclNode = XmlNode->new('common-acl');

  foreach my $param ( keys %params ) {
    if ( $param =~ /^DNS_Allow_Transfer/ ) {
      $aclNode->addChild( XmlNode->new( 'list-item', 'content' => $params{$param} ) );
    }
  }

  $dnsNode->addChild($aclNode);

  $parent->addChild($dnsNode);
}

sub addServerCertificate {
  my ( $self, $name, $cert, $csr, $ca_cert, $pvt_key, $default) = @_;

  my $parent = $self->{serverNode};
  if ( !defined($parent) ) {
    printToError('Error: addServerCertificate: empty parent node');
    return;
  }

  my $root = XmlNode->new('certificate');

  addUrlDecodedTextNode( $root, 'certificate-data', $cert ) if defined($cert);
  addUrlDecodedTextNode( $root, 'signing-request',  $csr ) if defined($csr);
  addUrlDecodedTextNode( $root, 'ca-certificate',   $ca_cert ) if defined($ca_cert);
  addUrlDecodedTextNode( $root, 'private-key', $pvt_key ) if defined($pvt_key);
  $root->setAttribute( 'name', $name );

  if ($default) {
  	$root->setAttribute('default', 'true');
  }

  $parent->getChild( 'certificates', 1 )->addChild($root);
}

sub addServerResellerTemplate {
  my ( $self, $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr ) = @_;

  my $root = $self->{serverNode};

  my $node = $self->makeTemplateNode('reseller-template', $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr );

  $root->getChild( 'account-templates', 1 )->addChild($node);
}

sub addServerClientTemplate {
  my ( $self, $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr ) = @_;

  my $root = $self->{serverNode};

  my $node = $self->makeTemplateNode('client-template', $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr );

  $root->getChild( 'account-templates', 1 )->addChild($node);
}

sub addServerDomainTemplate {
  my ( $self, $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr ) = @_;

  my $root = $self->{serverNode};

  my $node = $self->makeTemplateNode('domain-template', $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr );

  $root->getChild( 'account-templates', 1 )->addChild($node);
}

sub setServerSSO {
  my ( $self, $paramsPtr, $cert, $idpCert ) = @_;

  my %params = %{$paramsPtr};

  my $root = $self->{serverNode};

  my $ownership = defined $params{'admin_ownership'}
    && $params{'admin_ownership'} eq 'true' ? 'true' : 'false';
  my $enabled = defined $params{'sso_enabled'}
    && $params{'sso_enabled'} eq 'true' ? 'true' : 'false';

  my $ssoNode = XmlNode->new( 'sso-settings',
    'attributes' =>  { 'enabled' => $enabled }  );

  my @settings = (
    'sso_server',         'sso_relay',
    'sso_application_id', 'sso_idp_api_version',
    'sso_idp_id'
  );

  my %settings_map = (
    'sso_server'          => 'sso-server',
    'sso_relay'           => 'sso-relay',
    'sso_application_id'  => 'sso-application-id',
    'sso_idp_api_version' => 'sso-idp-api-version',
    'sso_idp_id'          => 'sso-idp-id'
  );

  foreach my $setting (@settings) {
    $ssoNode->addChild(
      XmlNode->new(
        $settings_map{$setting}, 'content' => $params{$setting}
      )
    ) if ( defined $params{$setting} );
  }

  if ( $cert )
  {
    $ssoNode->addChild( XmlNode->new( 'sso-certificate', 'content' => $cert ) );
  }

  if ( $idpCert ) {
    $ssoNode->addChild(
      XmlNode->new( 'sso-idp-cert', 'content' => $idpCert ) );
  }

  $root->addChild($ssoNode);
}

sub setServerSSOBranding {
  my ( $self, $brandingPtr ) = @_;

  my @branding = @{$brandingPtr};

  my $root = $self->{serverNode};

  my @ssoNodes = $root->getChildren('sso-settings');
  if ( scalar(@ssoNodes) < 1 ) {
    printToError("Error: setDomainCatchMail: there are no node 'sso-settings'");
    return;
  }
  my $ssoNode = $ssoNodes[0];

  my $ssoBrandingNode = XmlNode->new('sso-branding');
  if ( @branding ) {
    for my $ptrRow (@branding) {
      my $recordNode = XmlNode->new(
        'record',
        'children' => [
          XmlNode->new( 'idp-url', 'content' => $ptrRow->[1] ),
          XmlNode->new(
            'http-request-domain', 'content' => $ptrRow->[0]
          )
        ]
      );
      $ssoBrandingNode->addChild($recordNode);
    }
    $ssoNode->addChild($ssoBrandingNode);
  }
}

sub setServerAppVault {
  my ( $self ) = @_;

  my $root = $self->{serverNode};

  my @appVaultNodes = $root->getChildren('application-vault');
  if ( scalar(@appVaultNodes) > 0 ) {
    return $appVaultNodes[0];
  }

  my $appVaultNode = XmlNode->new('application-vault');

  $root->addChild($appVaultNode);

  return $appVaultNode;
}

sub setServerPackagesPool {
  my ( $self ) = @_;

  my $appVaultNode = $self->setServerAppVault();

  my @appPackagesPoolNodes = $appVaultNode->getChildren('sapp-packages-pool');
  if ( scalar(@appPackagesPoolNodes) > 0 ) {
    return $appPackagesPoolNodes[0];
  }

  my $appPackagesPoolNode = XmlNode->new('sapp-packages-pool');

  $appVaultNode->addChild($appPackagesPoolNode);

  return $appPackagesPoolNode;
}

sub addServerAppPackage {
  my ($self, $name, $version, $release, $distrib_path, $file_name) = @_;

  my $appPackagesPoolNode = $self->setServerPackagesPool();

  my $packageNode = XmlNode->new('sapp-package');
  $packageNode->addChild(  XmlNode->new( 'sapp-name', 'content' => $name ) );
  $packageNode->addChild(  XmlNode->new( 'sapp-version', 'content' => $version ) )  if ( $version );
  $packageNode->addChild( XmlNode->new( 'sapp-release', 'content' => $release ) )  if ( $release );

  if( $distrib_path && $file_name )
  {
    my $real_path = $distrib_path . "/" . $file_name;

    if ( -f $real_path ) {
      my $cid = $self->addAdminTar(
        'sapp-distrib',
        undef,
        "sapp-distrib." . generateDumpId(),
        "directory" => $distrib_path,
        "include"   => [$file_name]
      );
      $packageNode->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
    }
    else { return; }

    $appPackagesPoolNode->addChild($packageNode);
  }
}

sub setServerAppItemsPool {
  my ( $self ) = @_;

  my $appVaultNode = $self->setServerAppVault();

  my @appItemsNodes = $appVaultNode->getChildren('sapp-items-pool');
  if ( scalar(@appItemsNodes) > 0 ) {
    return $appItemsNodes[0];
  }

  my $appItemsNode = XmlNode->new('sapp-items-pool');

  $appVaultNode->addChild($appItemsNode);

  return $appItemsNode;
}

sub addServerAppItem {
  my ($self, $paramsPtr) = @_;

  my %params = %{$paramsPtr};

  my $appItemsPoolNode = $self->setServerAppItemsPool();

  my $appItemNode = XmlNode->new('sapp-item');
  $appItemNode->setAttribute( 'enabled',
    ( $params{'disabled'} eq 'true' ) ? 'false' : 'true' )
    if defined $params{'disabled'};
  my $appSpecNode = XmlNode->new('sapp-spec');
  $appSpecNode->addChild(
    XmlNode->new( 'sapp-name', 'content' => $params{'sapp_name'} ) );
  $appSpecNode->addChild(
    XmlNode->new(
      'sapp-version', 'content' => $params{'sapp_version'}
    )
  ) if defined $params{'sapp_version'};
  $appSpecNode->addChild(
    XmlNode->new(
      'sapp-release', 'content' => $params{'sapp_release'}
    )
  ) if defined $params{'sapp_release'};
  $appItemNode->addChild($appSpecNode);
  $appItemNode->addChild(
    XmlNode->new(
      'license-type',
      'content' => defined( $params{'license_type_id'} )
      ? $params{'license_type_id'}
      : "0"
    )
  );
  $appItemNode->addChild(
    XmlNode->new( 'shared', 'content' => $params{'shared'} ) )
    if defined $params{'shared'};
  $appItemNode->addChild(
    XmlNode->new( 'description', 'content' => $params{'description'} )
  ) if defined $params{'description'};
  $appItemNode->addChild(
    XmlNode->new(
      'instances-limit', 'content' => $params{'instances_limit'}
    )
  ) if defined $params{'instances_limit'};

  $appItemsPoolNode->addChild($appItemNode);
}

sub setServerAppLicensesPool {
  my ( $self ) = @_;

  my $appVaultNode = $self->setServerAppVault();

  my @appLicensesNodes = $appVaultNode->getChildren('sapp-licenses-pool');
  if ( scalar(@appLicensesNodes) > 0 ) {
    return $appLicensesNodes[0];
  }

  my $appLicensesNode = XmlNode->new('sapp-licenses-pool');

  $appVaultNode->addChild($appLicensesNode);

  return $appLicensesNode;
}

sub addServerAppLicense {
  my ($self, $keyNumber, $licenseType, $licenseText) = @_;

  my $licensePoolNode = $self->setServerAppLicensesPool();

  my $licenseNode = XmlNode->new(
    'sapp-license',
    'children' => [
      XmlNode->new( 'key-number',   'content' => $keyNumber ),
      XmlNode->new( 'license-type', 'content' => $licenseType ),
      XmlNode->new( 'license-text', 'content' => $licenseText )
    ]
  );

  $licensePoolNode->addChild($licenseNode);
}

sub setServerSBConfig {
  my ( $self, $configPtr ) = @_;

  my @config = @{$configPtr};

  my $serverNode = $self->{serverNode};

  my $sbNode = XmlNode->new('sb-config');
  for my $ptrHash (@config) {
    $sbNode->addChild(
      XmlNode->new(
        'sb-param',
        'children' => [
          XmlNode->new(
            'sb-param-name', 'content' => $ptrHash->{'param_name'}
          ),
          XmlNode->new(
            'sb-param-value', 'content' => $ptrHash->{'param_value'}
          )
        ]
      )
    );
  }
  $serverNode->addChild($sbNode);
}

sub setGLServerSettings {
  my ( $self, $paramsPtr) = @_;
 	 
  my %params = %{$paramsPtr};
 	 
  my $serverNode = $self->{serverNode};
	 
  my $glNode = XmlNode->new('grey-listing');
 	 
  $glNode->setAttribute('grey-interval', $params{'greyInterval'}) if defined $params{'greyInterval'};
  $glNode->setAttribute('expire-interval', $params{'expireInterval'}) if defined $params{'expireInterval'};
  $glNode->setAttribute('penalty-enabled', $params{'penaltyEnabled'}) if defined $params{'penaltyEnabled'};
  $glNode->setAttribute('penalty-interval', $params{'penaltyInterval'}) if defined $params{'penaltyInterval'};
  $glNode->setAttribute('enabled', $params{'enabled'}) if defined $params{'enabled'};
  $glNode->setAttribute('personal-conf', $params{'personal-conf'}) if defined $params{'personal-conf'};  

  my $wlNode = XmlNode->new('white-list');
 	 
  foreach my $wdomain (@{$params{'white_domains'}}) {
        $wlNode->addChild(XmlNode->new('list-item', 'content' => $wdomain));
  }
	 
  $glNode->addChild($wlNode);

  $serverNode->addChild($glNode);

}

sub setControlsVisibility {
  my ( $self, $paramsPtr ) = @_;

  my %params = %{$paramsPtr};

  my $serverNode = $self->{serverNode};
  my $controlsVisibilityNode = XmlNode->new('controls-visibility');
  
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-disabled-buttons', 'content' => HelpFuncs::checkValue($params{'hide_disabled_controls'},'false','true') ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-disabled-limits', 'content' => HelpFuncs::checkValue($params{'hide_unavailable_limits'},'false','true') ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-disabled-permissions', 'content' => HelpFuncs::checkValue($params{'hide_unavailable_permissions'},'false','true') ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-domain-registration-buttons', 'content' => HelpFuncs::negation(HelpFuncs::checkValue($params{'domain_registration'},'true','false')) ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-certificate-purchasing-buttons', 'content' => HelpFuncs::negation(HelpFuncs::checkValue($params{'cert_purchasing'},'true','false')) ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-extra-services-buttons', 'content' => HelpFuncs::negation(HelpFuncs::checkValue($params{'extras'},'true','false')) ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-mail-bouncing-controls', 'content' => HelpFuncs::negation(HelpFuncs::checkValue($params{'mail_bounce'},'true','false')) ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-newsfeeds', 'content' => HelpFuncs::negation(HelpFuncs::checkValue($params{'newsfeeds'},'true','false')) ) );
  $controlsVisibilityNode->addChild( XmlNode->new( 'hide-virtuozzo-promotion', 'content' => HelpFuncs::negation(HelpFuncs::checkValue($params{'promo_virtuozzo'},'true','false')) ) );
  
  
  $serverNode->getChild( 'interface-preferences', 1 )->addChild($controlsVisibilityNode);
}

sub setServerBackupSettings {
  my ( $self, $paramsPtr ) = @_;

  my %params = %{$paramsPtr};

  my $serverNode = $self->{serverNode};
  
  my ($lowPriority,$doNotCompress,$maxProcesses);
  
  $lowPriority = $params{'bu_nice'};
  $doNotCompress = $params{'bu_nozip'};
  $maxProcesses = $params{'max_bu_proc_number'};
  
  if ( defined $lowPriority or defined $doNotCompress or defined $maxProcesses ) {
  	my $serverBackupSettingsNode = XmlNode->new('backup-settings');
  	
  	$serverBackupSettingsNode->setAttribute( 'low-priority', $lowPriority)
			if defined $lowPriority and $lowPriority eq 'true';

  	$serverBackupSettingsNode->setAttribute( 'do-not-compress', $doNotCompress)
			if defined $doNotCompress and $doNotCompress eq 'true';

		$serverBackupSettingsNode->setAttribute( 'max-processes', $maxProcesses)
  		if defined $maxProcesses;

  	$serverNode->addChild($serverBackupSettingsNode);
  }
}

sub setServerPreferences {
  my ( $self, $paramsPtr ) = @_;

  my %params = %{$paramsPtr};

  my $serverNode = $self->{serverNode};

  my $serverPrefsNode = XmlNode->new('server-preferences');

  if ( defined $params{'forbid_create_dns_subzone'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'forbid-create-dns-subzone',
        'content' => $params{'forbid_create_dns_subzone'}
      )
    );
  }
  if ( defined $params{'force_db_user_prefix'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'force-db-user-prefix',
        'content' => $params{'force_db_user_prefix'}
      )
    );
  }

  if ( defined $params{'allow_siteapp_local_db'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'allow-siteapp-local-db',
        'content' => $params{'allow_siteapp_local_db'}
      )
    );
  }

  if ( defined $params{'db_user_length'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'db-user-length', 'content' => $params{'db_user_length'}
      )
    );
  }

  if ( defined $params{'hide_top_advertisement'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'hide-top-advertisement',
        'attributes' => { 'hide' => $params{'hide_top_advertisement'} }
      )
    );
  }
  if ( defined $params{'traffic_accounting'} ) {
    my $val = $params{'traffic_accounting'};
    if    ( $val == 1 ) { $val = 'in'; }
    elsif ( $val == 2 ) { $val = 'out'; }
    else                { $val = 'both'; }
    $serverPrefsNode->addChild(
      XmlNode->new(
        'traffic-direction', 'attributes' => { 'traffic' => $val }
      )
    );
  }
  if ( defined $params{'restart_apache_interval'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'restart-apache', 'content' => $params{'restart_apache_interval'}
      )
    );
  }
  if ( defined $params{'stat_ttl'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new( 'stat-keep', 'content' => $params{'stat_ttl'} ) );
  }
  if ( defined $params{'size_count_type'} ) {
    $serverPrefsNode->addChild(
      XmlNode->new(
        'disk-space-count-type',
        'attributes' => {
            'count-type' => $params{'size_count_type'} eq 'byte'
          ? 'byte'
          : 'block'
        }
      )
    );
  }

  my $duNode = XmlNode->new('disk-usage');
  $serverPrefsNode->addChild($duNode);
  $duNode->addChild( XmlNode->new('include-logs') )
    if defined $params{'include_logs'}
      and $params{'include_logs'} eq 'true';
  $duNode->addChild( XmlNode->new('include-databases') )
    if defined $params{'include_databases'}
      and $params{'include_databases'} eq 'true';
  $duNode->addChild( XmlNode->new('include-mailboxes') )
    if defined $params{'include_mailboxes'}
      and $params{'include_mailboxes'} eq 'true';
  $duNode->addChild( XmlNode->new('include-webapps') )
    if defined $params{'include_webapps'}
      and $params{'include_webapps'} eq 'true';
  $duNode->addChild( XmlNode->new('include-maillists') )
    if defined $params{'include_maillists'}
      and $params{'include_maillists'} eq 'true';
  $duNode->addChild( XmlNode->new('include-domaindumps') )
    if defined $params{'include_domaindumps'}
      and $params{'include_domaindumps'} eq 'true';

  $serverPrefsNode->addChild( XmlNode->new( 'force-db-prefix', 'content' => $params{'force_db_prefix'} ) ) if defined $params{'force_db_prefix'};
  $serverPrefsNode->addChild( XmlNode->new( 'multiple-session', 'content' => $params{'multiply_login'} ) ) if defined $params{'multiply_login'};
  $serverPrefsNode->addChild( XmlNode->new( 'lock-screen', 'content' => ($params{'disable_lock_screen'} eq 'true' ? 'false' : 'true' ) ) ) if defined $params{'disable_lock_screen'};

  $serverPrefsNode->addChild( XmlNode->new( 'aps-catalog-url', 'content' => $params{'apscatalog_url'} ) ) if defined $params{'apscatalog_url'};

  $serverNode->addChild($serverPrefsNode);
}

sub addRootAdmin {
  my ( $self, $id, $guid ) = @_;

  my $rootNode = $self->{root};
  if ( !defined($rootNode) ) {
    printToError('Error: addRootClient: empty parent node');
    return;
  }
  my $root = XmlNode->new( 'admin' );
  $self->{admin} = $root;
  $root->setAttribute( 'id', $id ) if $id;
  $root->setAttribute( 'guid', $guid ) if $guid;

  $rootNode->addChild($root);
}


### Client functions ##

sub addRootClient {
  my ( $self, $clientId, $clientPtr, $passwdPtr, $status, $ownerGuid ) = @_;

  my $rootNode = $self->{root};
  if ( !defined($rootNode) ) {
    printToError('Error: addRootClient: empty parent node');
    return;
  }

  my $root = $self->makeClientNode($clientId, $clientPtr, $passwdPtr, $status);
  $root->setAttribute( 'owner-guid', $ownerGuid ) if $ownerGuid;

  $rootNode->addChild($root);
  push @{$self->{roots}}, $root;
}

sub addRootReseller{
  my ( $self, $clientId, $clientPtr, $passwdPtr, $status ) = @_;

  my $rootNode = $self->{root};
  if ( !defined($rootNode) ) {
    printToError('Error: addRootClient: empty parent node');
    return;
  }

  my $root = $self->makeResellerNode($clientId, $clientPtr, $passwdPtr, $status);

  $rootNode->addChild($root);
  push @{$self->{roots}}, $root;
}

sub addResellerClient {
  my ( $self, $resellerId, $clientId, $clientPtr, $passwdPtr, $status ) = @_;

  my %client = %{$clientPtr};
  my %passwd = %{$passwdPtr};

  my $resellerNode = $self->{resellersNodes}->{$resellerId};
  if (!defined($resellerNode))
  {
    printToError('Error: addResellerClient: parent node is empty');
    return;
  }

  my $root = $self->makeClientNode($clientId, \%client, \%passwd, $status);
  my $resellerGuid = $resellerNode->getGuid();
  $root->setAttribute( 'owner-guid',  $resellerGuid) if $resellerGuid;

  $resellerNode->getChild( 'clients', 1 )->addChild($root);
  $self->regClientObjectBackupPath( $self->getResellersBackupPath( $resellerId ), $clientId, $clientPtr->{'login'} );
}

sub addAdminClient {
  my ( $self, $clientId, $clientPtr, $passwdPtr, $status ) = @_;

  my %client = %{$clientPtr};
  my %passwd = %{$passwdPtr};

  my $admin = $self->{admin};
  if (!defined($admin))
  {
    printToError('Error: addAdminClient: parent node is empty');
    return;
  }

  my $root = $self->makeClientNode($clientId, \%client, \%passwd, $status);
  my $adminGuid = $admin->getGuid();
  $root->setAttribute( 'owner-guid', $adminGuid ) if $adminGuid;

  $admin->getChild( 'clients', 1 )->addChild($root);
  $self->regClientObjectBackupPath( $self->getAdminBackupPath(), $clientId, $clientPtr->{'login'} );
}

sub addAdminReseller {
  my ( $self, $clientId, $clientPtr, $passwdPtr, $status ) = @_;

  my $admin = $self->{admin};
  if (!defined($admin))
  {
    printToError('Error: addAdminClient: parent node is empty');
    return;
  }

  my $root = $self->makeResellerNode($clientId, $clientPtr, $passwdPtr, $status);

  $admin->getChild( 'resellers', 1 )->addChild($root);
  $self->regResellersObjectBackupPath( $self->getAdminBackupPath(), $clientId, $clientPtr->{'login'} );
}

sub makeResellerNode{
  my ( $self, $clientId, $clientPtr, $passwdPtr, $status ) = @_;

  my $root = XmlNode->new( 'reseller' );
  $self->processClientNode( $root, $clientId, $clientPtr, $passwdPtr, $status, 1 );
  return $root;

}
sub makeClientNode {
  my ( $self, $clientId, $clientPtr, $passwdPtr, $status ) = @_;

  my $root = XmlNode->new( 'client' );
  $self->processClientNode( $root, $clientId, $clientPtr, $passwdPtr, $status, 0 );
  return $root;
}

sub processClientNode {
  my ( $self, $root, $clientId, $clientPtr, $passwdPtr, $status, $clientType ) = @_;

 $root->getChild( 'preferences', 1 );
 $root->getChild( 'properties', 1 );
 $root->getChild( 'limits-and-permissions', 1 );

  my %client = %{$clientPtr};
  my %passwd = %{$passwdPtr};

  $root->setAttribute( 'id', $clientId );

  if ( defined $client{'uid'} ) {
    $root->setAttribute( 'uid', $client{'uid'} );
    $root->setAttribute( 'ownership', $client{'ownership'} eq 'true' ? 'true' : 'false' );
  }

  $root->setAttribute( 'guid', $client{'guid'} ) if ( defined $client{'guid'} );

  my ( $attribute, $field, $name );

  if ( exists $client{'locale'} && $client{'locale'} ) {
    while ( ( $name, $field ) = each(%locale_map) ) {
      if ( $client{'locale'} eq $name ) {
        $client{'locale'} = $field;
      }
    }
  }

  while ( ( $attribute, $field ) = each(%clientsAttribute) ) {
    if ( exists( $client{$field} ) ) {
      unless ( defined( $client{$field} ) ) {
        $client{$field} = '';
      }
      $root->setAttribute( $attribute, $client{$field} );
    }
  }

  my $item;
  my $props = $root->getChild( 'properties', 1 );
  $item =  makePasswordNode( $passwd{'password'}, normalizePasswordType( $passwd{'type'} ) );
  $props->addChild($item);

  $props->addChild( Status::make($status) );

  $self->{clientNodes}->{$clientId} = $root;
  $self->{resellersNodes}->{$clientId} = $root if $clientType;

}

sub finishClient() {
  my ( $self, $clientId ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: finishClient: empty parent node');
    return;
  }

  #TODO: uncomment
  #$root->ReleaseCode();
}

sub setClientPinfo {
  my ( $self, $clientId, $clientPtr ) = @_;

  my %client = %{$clientPtr};

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: setClientPinfo: empty parent node');
    return;
  }

  my ( $attribute, $field, $name );

  my $prefs = $root->getChild( 'preferences', 1 );
  while ( ( $name, $field ) = each(%clientsInfo) ) {
    if ( exists( $client{$field} ) && $client{$field} ) {
      $prefs->addChild(
        XmlNode->new(
          'pinfo',
          'content'    => $client{$field},
          'attributes' => { 'name' => $name }
        )
      );
    }
  }
}

sub setClientLocale {
  my ( $self, $clientId, $locale ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: setClientLocale: empty parent node');
    return;
  }

  $root->getChild( 'preferences', 1 )->addChild(
    XmlNode->new(
      'pinfo',
      'content'    => $locale,
      'attributes' => { 'name' => 'locale' }
    )
  );
}

sub setClientMaxButtonLength {
  my ( $self, $clientId, $length ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: setClientMaxButtonLength: empty parent node');
    return;
  }

  $root->setAttribute( 'max-button-length', $length );
}

sub setClientSkin {
  my ( $self, $clientId, $skin ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: setClientSkin: empty parent node');
    return;
  }

  $root->setAttribute( 'skin', $skin );
}

sub setClientLockScreen {
  my ( $self, $clientId ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: setClientLokScreen: empty parent node');
    return;
  }

  $root->setAttribute( 'lock-screen', 'false' );
}

sub addClientLimit {
  my ( $self, $clientId, $name, $value ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: addClientLimit: empty parent node');
    return;
  }

  $self->insertLimitNode( $root, $name, $value );
}

sub addClientPermission {
  my ( $self, $clientId, $name, $value ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: addClientPermission: empty parent node');
    return;
  }

  $self->makePermissionNode( $root->getChild( 'limits-and-permissions', 1 ), $name, $value );
}

sub addClientDomainSkeleton {
  my ( $self, $clientId, $path, $arc_path ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: addClientDomainSkeleton: empty parent node');
    return;
  }

  my $dumpFile = $self->addClientTar( 'skeleton', $clientId, $arc_path, 'directory' => $path );
  $root->getChild( 'content', 1, 1 )->addChild( $dumpFile) if $dumpFile;
}

sub addServerSkeleton {
  my ( $self, $path, $arc_path) = @_;

  my $root = $self->{serverNode};

  my $dumpFile = $self->addAdminTar('skeleton', undef, $arc_path, 'directory' => $path);

  $root->getChild('content', 1, 1)->addChild($dumpFile) if $dumpFile;
}

sub addClientIps {
  my ( $self, $clientId, $ipsPtr ) = @_;

  my %ips = %{$ipsPtr};

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: addClientIps: empty parent node');
    return;
  }

  my $ip_pool = XmlNode->new('ip_pool');
  $root->addChild($ip_pool);

  while( my($ip,$iptype) = each(%{$ipsPtr}) ){
   $ip_pool->addChild( $self->makeIpNode($ip, $iptype) );
  }
}

sub setClientSappPool {
  my ( $self, $clientId ) = @_;

  my $root = $self->{clientNodes}->{$clientId};
  if ( !defined($root) ) {
    printToError('Error: setClientSappPool: empty parent node');
    return;
  }

  my $sapp_pool = XmlNode->new('sapp-pool');

  $root->addChild($sapp_pool);
}

sub addClientSapp {
  my ( $self, $clientId, $ptrHash ) = @_;

  my $root = $self->{clientNodes}->{$clientId};

  my @pools = $root->getChildren('sapp-pool');
  if ( scalar(@pools) < 1 ) {
    $self->setClientSappPool($clientId);
    @pools = $root->getChildren('sapp-pool');
  }

  my $sapp_pool = $pools[0];

  $self->makeSiteAppItemNode( $sapp_pool, $ptrHash );
}


sub setClientTraffic{
  my ( $self, $clientId, $trafficValue ) = @_;

  my $parent = $self->{clientNodes}->{$clientId};
  if ( !defined($parent) ) {
    printToError('Error: setClientTraffic: empty parent node');
    return;
  }

  $parent->addChild( XmlNode->new( 'traffic', 'content' => $trafficValue ) );
}

sub addResellerDomainTemplate {
  my $self = shift;
  $self->addClientDomainTemplate( @_ );
}

sub addClientDomainTemplate {
  my ( $self, $clientId, $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr ) = @_;

  my $root = $self->{clientNodes}->{$clientId};

  my $node = $self->makeTemplateNode('domain-template', $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr );

  $root->getChild( 'preferences', 1 )->addChild($node);
}

sub addResellerClientTemplate {
  my ( $self, $resellerId, $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr ) = @_;

  my $root = $self->{resellersNodes}->{$resellerId};

  my $node = $self->makeTemplateNode('client-template', $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr );

  $root->getChild( 'preferences', 1 )->addChild($node);
}

sub addClientCustomButton {
    my ( $self, $clientId, $id, $optionsPtr, $customButtonsDir, $icon ) = @_;

    my $parent = $self->{clientNodes}->{$clientId};
    if ( !defined($parent) ) {
      printToError('Error: setClientCustomButton: empty parent node');
      return;
    }

    my $node = $self->makeCustomButtonNode( 'clients', $clientId, $id, $optionsPtr, $customButtonsDir,
      $icon );

    $parent->getChild( 'preferences', 1 )->addChild($node);
}

sub setClientSiteBuilder {
    my ( $self, $clientId, $sbClientsPtr, $enabled) = @_;

    my @sbClients = @{$sbClientsPtr};

    my $parent = $self->{clientNodes}->{$clientId};
    if ( !defined($parent) ) {
      printToError('Error: setClientSiteBuilder: empty parent node');
      return;
    }

    my $sbNode = XmlNode->new('sb-reseller');

    if (defined($enabled)) {
      $sbNode->setAttribute('enabled', $enabled);
	}

    for  my $ptrHash (@sbClients) {
      $sbNode->addChild(
        XmlNode->new(
          'sb-client-login', 'content' => $ptrHash->{'sb_client_login'}
        )
      );
      $sbNode->addChild(
        XmlNode->new(
          'sb-reseller-id', 'content' => $ptrHash->{'sb_reseller_id'}
        )
      );
    }

    $parent->getChild( 'properties', 1 )->addChild($sbNode);
}

### Domain functions ##

sub addRootDomain {
    my ( $self, $domainId, $domainName, $domainAsciiName, $domainPtr, $ownerGuid ) = @_;
    $domainAsciiName = $domainName if not $domainAsciiName;

    my $rootNode = $self->{root};
    if ( !defined($rootNode) ) {
      printToError('Error: addRootDomain: empty parent node');
      return;
    }

    my $root = $self->makeDomainNode($domainId, $domainName, $domainPtr);
    $root->setAttribute( 'owner-guid', $ownerGuid ) if $ownerGuid;

    $rootNode->addChild($root);
    push @{$self->{roots}}, $root;
}

sub addAdminDomain {
    my ( $self, $domainId, $domainName, $domainAsciiName, $domainPtr ) = @_;
    $domainAsciiName = $domainName if not $domainAsciiName;

    my $admin = $self->{admin};
    if ( !defined($admin) ) {
      printToError('Error: addAdminDomain: empty parent node');
      return;
    }

    my $root = $self->makeDomainNode($domainId, $domainName, $domainPtr);
    my $adminGuid = $admin->getGuid();
    $root->setAttribute( 'owner-guid', $admin->getGuid() ) if $adminGuid;

    $admin->getChild( 'domains', 1 )->addChild($root);
    $self->regDomainObjectBackupPath( $self->getAdminBackupPath(), $domainId, $domainAsciiName );

}

sub addResellerDomain {
    my ( $self, $resellerId, $domainId, $domainName, $domainAsciiName, $domainPtr ) = @_;
    $domainAsciiName = $domainName if not $domainAsciiName;

    my $resellerNode = $self->{resellersNodes}->{$resellerId};
    if ( !defined($resellerNode) ) {
      printToError('Error: addResellerDomain: empty parent node');
      return;
    }

    my $root = $self->makeDomainNode($domainId, $domainName, $domainPtr);
    my $resellerGuid = $resellerNode->getGuid();
    $root->setAttribute( 'owner-guid',  $resellerGuid) if $resellerGuid;

    $resellerNode->getChild( 'domains', 1 )->addChild($root);
    $self->regDomainObjectBackupPath( $self->getResellersBackupPath( $resellerId ), $domainId, $domainAsciiName );
}

sub addClientDomain {
    my ( $self, $clientId, $domainId, $domainName, $domainAsciiName, $domainPtr ) = @_;
    $domainAsciiName = $domainName if not $domainAsciiName;

    my $clientNode = $self->{clientNodes}->{$clientId};
    if ( !defined($clientNode) ) {
      printToError('Error: addClientDomain: empty parent node');
      return;
    }

    my $root = $self->makeDomainNode($domainId, $domainName, $domainPtr);
    my $clientGuid = $clientNode->getGuid();
    $root->setAttribute( 'owner-guid', $clientGuid ) if $clientGuid;

    $clientNode->getChild( 'domains', 1 )->addChild($root);
    $self->regDomainObjectBackupPath( $self->getClientsBackupPath( $clientId ), $domainId, $domainAsciiName );
}

sub makeDomainNode {
    my ( $self, $domainId, $domainName, $domainPtr ) = @_;

    unless ( ref($domainPtr) =~ /HASH/ ) {
      printToError("Error: makeDomainNode: domainPtr is not ref hash");
      return undef;
    }

    my %domain = %{$domainPtr};

    my $root = XmlNode->new( 'domain', 'attributes' => {
        'name' => $domainName,
        'id'   => $domainId
    } );

    $root->getChild( 'preferences', 1 );
    $root->getChild( 'properties', 1 );
    $root->getChild( 'limits-and-permissions', 1 );

    $root->setAttribute( 'cr-date', $domain{'cr_date'} ) if exists $domain{'cr_date'};
    $root->setAttribute( 'guid', $domain{'guid'} ) if defined $domain{'guid'};

    $self->{domainNodes}->{$domainId} = $root;

    return $root;
}

sub finishDomain() {
  my ( $self, $domainId) = @_;

  my $root = $self->{domainNodes}->{$domainId};
  if ( !defined($root) ) {
    printToError('Error: finishDomain: empty parent node');
    return;
  }

  #TODO: uncomment
  #$root->ReleaseCode();
}

sub setDomainWwwStatus {
    my ( $self, $domainId, $status ) = @_;

    my $parent = $self->{domainNodes}->{$domainId};

    $parent->setAttribute( 'www', $status );
}

sub setDomainIP {
    my ( $self, $domainId, $ip, $iptype ) = @_;

    my $root = $self->{domainNodes}->{$domainId};

    $root->getChild( 'properties', 1 )->addChild( $self->makeIpNode( $ip, $iptype ) );
}

sub setDomainStatus {
    my ( $self, $domainId, $status ) = @_;

    my $root = $self->{domainNodes}->{$domainId};

    $root->getChild( 'properties', 1 )->addChild( Status::make($status) );
}

sub addDomainDatabase {
    my ( $self, $dbId, $dbServerId, $domainId, $dbName, $dbType, $optionalPtr, $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr ) = @_;

    $self->regDatabaseObjectBackupPath( $self->getDomainsBackupPath( $domainId ), $dbId, $dbName, $dbServerId );
    my $parent = $self->{domainNodes}->{$domainId};

    my $node = $self->makeDatabaseNode( $dbId, $dbName, $dbType, 'domain', $domainId, $optionalPtr,
                                        $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr, $domainId  );
    my $parentGuid = $parent->getGuid();
    $node->setAttribute( 'guid', $self->getDatabaseGuid( $parentGuid, $dbId ) ) if $parentGuid;
    $node->setAttribute( 'owner-guid', $parentGuid ) if $parentGuid;

    $parent->getChild( 'databases', 1 )->addChild($node);
}

sub setDomainDefault {
    my ( $self, $domainId ) = @_;

    my $root = $self->{domainNodes}->{$domainId};

    $root->getChild( 'preferences', 1 )->addChild( XmlNode->new('default-domain') );
}

sub addDomainLimit {
    my ( $self, $domainId, $name, $value ) = @_;

    my $root = $self->{domainNodes}->{$domainId};
    if ( !defined($root) ) {
      printToError('Error: addDomainLimit: empty parent node');
      return;
    }

    $self->insertLimitNode( $root, $name, $value );
}

sub setDomainMailService {
    my ( $self, $domainId, $status ) = @_;

    my $root = $self->{domainNodes}->{$domainId};
    if ( !defined($root) ) {
      printToError('Error: setDomainMailService: empty parent node');
      return;
    }

    my $mailServiceStatus = Status::make($status);

    my $rootMail = XmlNode->new('mailsystem');
    $rootMail->getChild( 'properties', 1 )->addChild($mailServiceStatus);

    $root->addChild($rootMail);
}

sub addDomainCustomButton {
    my ( $self, $domainId, $id, $optionsPtr, $customButtonsDir, $icon ) = @_;

    my $parent = $self->{domainNodes}->{$domainId};
    if ( !defined($parent) ) {
      printToError('Error: addDomainCustomButton: empty parent node');
      return;
    }

    my $node = $self->makeCustomButtonNode( 'domains', $domainId, $id, $optionsPtr, $customButtonsDir,
      $icon );

    $parent->getChild( 'preferences', 1 )->addChild( $node);
}

sub setDomainCatchMail {
    my ( $self, $domainId, $catchAllAddr ) = @_;

    my $parent = $self->{domainNodes}->{$domainId};
    if ( !defined($parent) ) {
      printToError('Error: setDomainCatchMail: empty parent node');
      return;
    }

    my $mailService = $parent->getChild('mailsystem');
    if ( not $mailService ) {
      printToError("Error: setDomainCatchMail: there are no node 'mailsystem'");
      return;
    }

    $mailService->getChild( 'preferences', 1 )->addChild(XmlNode->new( 'catch-all', 'content' => $catchAllAddr ) );
}

sub setDomainKeysDomainSupport {
    my ( $self, $domainId, $domainName, $state, $privateKeyPath,
      $publickKey ) = @_;

    my $parent = $self->{domainNodes}->{$domainId};
    if ( !defined($parent) ) {
      printToError('Error: setDomainKeysDomainSupport: empty parent node');
      return;
    }

    my $mailService = $parent->getChild('mailsystem');
    if ( not $mailService ) {
      printToError("Error: setDomainKeysDomainSupport: there are no node 'mailsystem'");
      return;
    }

    my $domainKeysNode = XmlNode->new( 'domain-keys',
      'attributes' => { 'state' => $state } );

    if ($state) {
      my $privateKey = $self->addDomainTar( 'key', $domainId,
        "privatekey",
        'directory' => $privateKeyPath,
        'include'   => ['default']
      );

      $domainKeysNode->getChild( 'content', 1, 1 )->addChild( $privateKey ) if $privateKey;

      $domainKeysNode->setAttribute( 'public-key', $publickKey ) if ( defined($publickKey) && $publickKey );
    }

    $mailService->getChild( 'preferences', 1 )->addChild($domainKeysNode);
}

sub setDomainWebMail{
  my ( $self, $domainId, $webmail ) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
   if ( !defined($parent) ) {
     printToError('Error: setDomainWebMail: empty parent node');
     return;
   }

  my $mailService = $parent->getChild('mailsystem');
  if ( not $mailService ) {
     printToError("Error: setDomainKeysDomainSupport: there are no node 'mailsystem'");
     return;
  }

  my $node = XmlNode->new( 'web-mail' );
  $node->setText( $webmail );
  $mailService->getChild( 'preferences', 1 )->addChild( $node );
}

sub setDomainGLSupport {
    my ( $self, $domainId, $state ) = @_;
    my $parent = $self->{domainNodes}->{$domainId};
    my $mailsystem = $parent->getChild('mailsystem');
    my $node = XmlNode->new( 'grey-listing' );
    $node->setText($state);
    $mailsystem->getChild('preferences', 1)->addChild($node);
}

sub setDomainDnsZone {
    my ( $self, $domainId, $paramsPtr, $recordsPtr ) = @_;

    my $parent = $self->{domainNodes}->{$domainId};
    if ( !defined($parent) ) {
      printToError('Error: setDomainDnsZone: empty parent node');
      return;
    }

    $self->makeDnsZone($parent->getChild( 'properties', 1 ), $paramsPtr, $recordsPtr);
}

sub setDomainDnsOld {
    my ( $self, $domainId, $enabled, $type, $masterIp, $recordsPtr ) = @_;

    my @records = @{$recordsPtr};

    my $parent = $self->{domainNodes}->{$domainId};
    if ( !defined($parent) ) {
      printToError('Error: setDomainDnsOld: empty parent node');
      return;
    }
	$parent = $parent->getChild( 'properties', 1 );
    if ( !defined($parent) ) {
      printToError('Error: setDomainDnsOld: empty parent node');
      return;
    }

    my $dnsNode = XmlNode->new('dns-zone');

    $dnsNode->setAttribute(
      'type',
      ( !$type or ( $type eq 'master' ) ) ? 'master' : 'slave'
    );

	$dnsNode->addChild(Status::make( $enabled eq 'true' ? 0 : 16 ));

    for my $hash (@records) {
      my $dnsrec = $self->makeDnsRecord($hash);
      if ($dnsrec) {
        $dnsNode->addChild($dnsrec);
      }
      $hash = undef;
    }

    # Add old master record

    my $sql = "SELECT * FROM dns_masters WHERE dom_id = '$domainId'";

    if ( defined($masterIp) ) {
      $dnsNode->addChild( makeOldMasterRec($masterIp) );
    }

    $parent->addChild($dnsNode);
}

sub addDomainAlias {
  my ( $self, $domainId, $aliasPtr, $dnsParamsPtr, $dnsRecordsPtr) = @_;

  my %alias = %{$aliasPtr};

  my $parent = $self->{domainNodes}->{$domainId};
  if ( !defined($parent) ) {
    printToError('Error: addDomainAlias: empty parent node');
    return;
  }

  my $aliasNode = XmlNode->new(
    'domain-alias',
    'attributes' => {
      'name' => $alias{'displayName'},
      'mail' => $alias{'mail'},
      'web'  => $alias{'web'},
      'dns'  => defined $alias{'dns'} ? $alias{'dns'} : 'false'
    },
    'children' => [ Status::make( $alias{'status'} ) ]
  );

  $aliasNode->setAttribute( 'tomcat', $alias{'tomcat'} )
    if defined $alias{'tomcat'};

  # DNS zone

  if ( $alias{'dns_zone_id'} != 0 ) {
    $self->makeDnsZone( $aliasNode, $dnsParamsPtr, $dnsRecordsPtr );
  }

  $parent->getChild( 'preferences', 1 )->addChild($aliasNode);
}

sub setDomainMailLists {
  my ( $self, $domainId, $status) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if ( !defined($parent) ) {
    printToError('Error: setDomainMaillists: empty parent node');
    return;
  }

  my $mlNode = XmlNode->new('maillists');
  $mlNode->addChild(Status::make($status));

  $parent->addChild($mlNode);
}

sub addDomainMailList {
  my ( $self, $domainId, $name, $passwd, $state, $ownersPtr, $membersPtr, $content_path, $content_file) = @_;

  my @owners = @{$ownersPtr};
  my %members = %{$membersPtr};

  my $parent = $self->{domainNodes}->{$domainId};
  if ( !defined($parent) ) {
    printToError('Error: addDomainMaillist: empty parent node');
    return;
  }

  my @mailListServs = $parent->getChildren('maillists');
  if ( scalar(@mailListServs) < 1 ) {
    printToError('Error: addDomainMailList: there are no node "maillists"');
    return;
  }
  my $mailListService = $mailListServs[0];

  my $node = XmlNode->new(
    'maillist',
    'attributes' => { 'name', $name },
    'children'   => [ Status::make($state) ]
  );

  for my $listOwner (@owners) {
    $node->addChild( XmlNode->new( 'owner', 'content' => $listOwner ) );
  }

  $node->addChild( makePasswordNode( $passwd, 'encrypted') );

  for my $memberEmail ( keys %members ) {
    my $member = XmlNode->new( 'recipient', 'content' => $memberEmail );
    if ( $members{$memberEmail} ne "" ) {
      $member->setAttribute( 'fullname', $members{$memberEmail} );
    }
    $node->addChild($member);
  }

  if ( -f "$content_path/$content_file" ) {
    my $cid = $self->addDomainMailListTar( 'mailbox', $domainId,
      "$name.mlist",
      'directory'       => $content_path,
      'follow_symlinks' => 1,
      'include' => [$content_file]
    );
    $node->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
  }

  $mailListService->addChild($node);
}

sub setDomainTraffic ($ $) {
  my ( $self, $domainId, $trafficValue) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if ( !defined($parent) ) {
    printToError('Error: setDomainTraffic: empty parent node');
    return;
  }

  $parent->addChild( XmlNode->new( 'traffic', 'content' => $trafficValue ) );
}

sub addDomainCertificate {
  my ( $self, $domainId, $name, $cert, $csr, $ca_cert, $pvt_key) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if ( !defined($parent) ) {
    printToError('Error: addDomainCertificate: empty parent node');
    return;
  }

  my $root = XmlNode->new('certificate');

  addUrlDecodedTextNode( $root, 'certificate-data', $cert ) if defined($cert);
  addUrlDecodedTextNode( $root, 'signing-request',  $csr ) if defined($csr);
  addUrlDecodedTextNode( $root, 'ca-certificate',   $ca_cert ) if defined($ca_cert);
  addUrlDecodedTextNode( $root, 'private-key', $pvt_key ) if defined($pvt_key);
  $root->setAttribute( 'name', $name );

  $parent->getChild('certificates', 1 )->addChild($root);
}

sub setDomainTomcat {
  my ( $self, $domainId, $status) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if (!defined($parent) ) {
    printToError('Error: setDomainTomcat: empty parent node');
    return;
  }

  my $tomcatNode = XmlNode->new('tomcat');
  $tomcatNode->addChild( Status::make( $status ) );
  $parent->addChild($tomcatNode);
}

sub addDomainTomcatWebApp {
  my ( $self, $domainId, $domainName, $name, $status, $path) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if (!defined($parent) ) {
    printToError('Error: addDomainTomcatWebApp: empty parent node');
    return;
  }

  my @tomcatServs = $parent->getChildren('tomcat');
  if ( scalar(@tomcatServs) < 1 ) {
    printToError('Error: addDomainTomcatWebApp: there are no node "tomcat"');
    return;
  }
  my $tomcatService = $tomcatServs[0];

  my $item = XmlNode->new(
    'webapp',
    'attributes' => { 'name' => $name },
    'children'   => [ Status::make($status) ]
  );

  my $warFile = "$name.war";

  my $cid = $self->addDomainTar( 'web-app', $domainId,
    "$name.webapp",
    "directory" => $path,
    "include"   => [$warFile]
  );
  $item->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;

  $tomcatService->addChild($item);
}

sub addDomainUser {
  my ( $self, $domainId, $passwd, $passwdType, $state, $locale, $paramsPtr) = @_;

  my %params = %{$paramsPtr};

  my $parent = $self->{domainNodes}->{$domainId};
  if (!defined($parent) ) {
    printToError('Error: addDomainUser: empty parent node');
    return;
  }

  my $root = XmlNode->new('domainuser');

  if ( $state ) {
    $root->setAttribute( 'status', $state );
  }

  if (defined($passwd)) {
      $root->addChild( makePasswordNode( $passwd, $passwdType ) );
  }
  while ( my ( $xmlName, $fieldName ) = each %domainUserAttribute ) {
    if ( $params{$fieldName} ) {
      $root->setAttribute( $xmlName, $params{$fieldName} );
    }
  }

  while ( my ( $xmlName, $fieldName ) = each %domainUserInfo ) {
    my $value = $params{$fieldName};
    if ($value) {
      $value = normalizeCountry($value) if $fieldName eq 'country';
      $root->addChild(
        XmlNode->new(
          'pinfo',
          'content'    => $value,
          'attributes' => { 'name' => $xmlName }
        )
      );
    }
  }

  $root->addChild(
    XmlNode->new(
      'pinfo',
      'content'    => $locale,
      'attributes' => { 'name' => 'locale' }
    )
  ) if defined $locale;


  $parent->addChild($root);
}

sub addDomainPermission {
  my ( $self, $domainId, $name, $value ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @domainUsers = $root->getChildren('domainuser');
  if ( scalar(@domainUsers) < 1 ) {
    printToError('Error: addDomainPermission: there are no node "domainuser"');
    return;
  }

  $self->makePermissionNode( $domainUsers[0], $name, $value );
}

sub setDomainMaxButtonLength {
  my ( $self, $domainId, $length ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @domainUsers = $root->getChildren('domainuser');
  if ( scalar(@domainUsers) < 1 ) {
    printToError('Error: setDomainMaxButtonLength: there are no node "domainuser"');
    return;
  }

  $domainUsers[0]->setAttribute( 'max-button-length', $length );
}

sub setDomainSkin {
  my ( $self, $domainId, $skin ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @domainUsers = $root->getChildren('domainuser');
  if ( scalar(@domainUsers) < 1 ) {
    printToError('Error: setDomainSkin: there are no node "domainuser"');
    return;
  }

  $domainUsers[0]->setAttribute( 'skin', $skin );
}

sub setDomainLockScreen {
  my ( $self, $domainId ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @domainUsers = $root->getChildren('domainuser');
  if ( scalar(@domainUsers) < 1 ) {
    printToError('Error: setDomainLockScreen: there are no node "domainuser"');
    return;
  }

  $domainUsers[0]->setAttribute( 'lock-screen', 'false' );
}

sub setDomainPhosting {
  my ( $self, $domainId, $paramsPtr, $sysuserPtr, $scriptingPtr ) = @_;

  my %params = %{$paramsPtr};
  my %sysuser = %{$sysuserPtr};
  my %scripting = %{$scriptingPtr};

  my $parent = $self->{domainNodes}->{$domainId};

  my $root = XmlNode->new('phosting');
  my $prefs = $root->getChild( 'preferences', 1 );
  my $parentGuid = $parent->getGuid();
  $root->setAttribute( 'guid', $self->getPhostingGuid( $parentGuid , $domainId ) ) if $parentGuid;
  $root->setAttribute( 'owner-guid', $parentGuid ) if $parentGuid;

  ## %params ##

  $root->setAttribute( 'shared-content', $params{'shared-content'} ) if defined($params{'shared-content'});

  while ( my ( $xmlName, $fieldName ) = each(%hostingAttribute) ) {
    # MySQL allows 'true', 'false' and '' in enums
    if ( defined( $params{$fieldName} )
      and $params{$fieldName} eq "true" )
    {
      $root->setAttribute( $xmlName, 'true' );
    }
  }

  if ( defined( $params{'webstat'} ) and $params{'webstat'} ) {
    $root->setAttribute( 'webstat', $params{'webstat'} );
  }

  if ( defined( $params{'https'} ) and $params{'https'} ) {
    $root->setAttribute( 'https', $params{'https'} );
  }

  if ( defined($params{'errdocs'}) and $params{'errdocs'} eq 'true' ) {
    $root->setAttribute( 'errdocs', 'true' );
  }

  ## %sysuser ##

  if (%sysuser)
  {
    $prefs->addChild($self->makeSysUser(\%sysuser));
  }

  ## %scripting ##

  my $scripringItem = XmlNode->new('scripting');
  while ( my ( $xmlName, $value ) = each(%scripting) ) {
      $scripringItem->setAttribute( $xmlName, $value );
  }
  $root->setAttribute( 'wu_script', $params{'wu_script'} ) if exists $params{'wu_script'};

  $root->getChild( 'limits-and-permissions', 1 )->addChild($scripringItem);


  $parent->addChild($root);

  $self->regPhostingObjectBackupPath( $self->getDomainsBackupPath( $domainId ), $domainId );
  $self->{phostingNodes}->{$domainId} = $root;

}

sub setDomainPhostingFullContent {
  my ( $self, $domainId, $domainName, $path) = @_;
  my $root = $self->{domainNodes}->{$domainId};

  my $phostings = $root->getChild('phosting');
  my $cid_content = $self->addDomainTar( 'vhost', $domainId,
    "vhost",
    "directory" => $path
  );
  $phostings->getChild( 'content', 1, 1 )->addChild( $cid_content ) if $cid_content;
  @{$self->{domainVhost}->{$domainId}} = ( $cid_content, $path )  if $cid_content;
}

sub setDomainMailNamesContent{
    my ( $self, $domainId, $domainAsciiName, $path ) = @_;

    my $root = $self->{domainNodes}->{$domainId};
    my $mailContent = $self->addDomainTar( 'domainmail', $domainId,
        "mn",
        'directory' => $path
      );

   $root->getChild( 'mailsystem' )->getChild( 'content', 1, 1 )->addChild( $mailContent ) if $mailContent;
   @{$self->{domainMailContent}->{$domainId}} = ( $mailContent, $path ) if $mailContent;
}

sub setDomainMailListContent{
    my ( $self, $domainId, $domainAsciiName, $path, $includePtr ) = @_;

    my $root = $self->{domainNodes}->{$domainId};
    my $mailContent = $self->addDomainTar( 'domainmaillist', $domainId,
        "ml",
        'directory' => $path,
        'include' => $includePtr,
        'follow_symlinks' => 1
      );

   $root->getChild( 'maillists' )->getChild( 'content', 1, 1 )->addChild( $mailContent ) if $mailContent;
   @{$self->{domainMailListContent}->{$domainId}} = ( $mailContent, $path ) if $mailContent;
}


sub setDomainPhostingHttpdocsContent {
  my ( $self, $domainId, $domainName, $path, $exclude_httpdocs_files_ptr ) = @_;

  my @exclude_httpdocs_files = @{$exclude_httpdocs_files_ptr};

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingHttpdocsContent: there are no node "phosting"');
    return;
  }

  my $cid_docroot = $self->addPhostingTar( 'docroot', $domainId,
    "httpdocs",
    "directory" => $path,
    "exclude"   => \@exclude_httpdocs_files
  );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_docroot ) if $cid_docroot;
}

sub setDomainPhostingHttpsdocsContent {
  my ( $self, $domainId, $domainName, $path, $exclude_httpsdocs_files_ptr ) = @_;

  my @exclude_httpsdocs_files = @{$exclude_httpsdocs_files_ptr};

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingHttpsdocsContent: there are no node "phosting"');
    return;
  }

  my $cid_docroot_ssl = $self->addPhostingTar( 'docroot_ssl', $domainId,
    "httpsdocs",
    "directory" => $path,
    "exclude"   => \@exclude_httpsdocs_files
  );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_docroot_ssl ) if $cid_docroot_ssl;
}

sub setDomainPhostingCgibinContent {
  my ( $self, $domainId, $domainName, $path, $exclude_cgi_bin_files_ptr ) = @_;

  my @exclude_cgi_bin_files = @{$exclude_cgi_bin_files_ptr};

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_cgi = $self->addPhostingTar( 'cgi', $domainId,
    "cgi-bin",
    "directory" => $path,
    "exclude"   => \@exclude_cgi_bin_files
  );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_cgi ) if $cid_cgi;
}

sub setDomainPhostingWebstatContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_webstat = $self->addPhostingTar( 'webstat', $domainId, "webstat", "directory" => "$path" );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_webstat ) if $cid_webstat;
}

sub setDomainPhostingWebstatSslContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_webstat_ssl = $self->addPhostingTar( 'webstat_ssl', $domainId, "webstat-ssl",  "directory" => $path );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_webstat_ssl ) if $cid_webstat_ssl;
}

sub setDomainPhostingFtpstatContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_ftpstat = $self->addPhostingTar( 'ftp_stat', $domainId, "ftpstat", "directory" => $path);
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_ftpstat ) if $cid_ftpstat;
}

sub setDomainPhostingErrdocsContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_error_docs = $self->addPhostingTar( 'error_docs', $domainId, "error-docs",
    "directory" => $path );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_error_docs ) if $cid_error_docs;
}

sub setDomainPhostingPrivateContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_private = $self->addPhostingTar( 'private', $domainId, "private",  "directory" => $path );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_private ) if $cid_private;
}

sub setDomainPhostingLogsContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_logs =  $self->addPhostingTar( 'logs', $domainId, "logs", "directory" => $path );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_logs ) if $cid_logs;
}

sub setDomainPhostingAnonFtpstatContent {
  my ( $self, $domainId, $domainName, $path ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingCgibinContent: there are no node "phosting"');
    return;
  }

  my $cid_anon_ftpstat = $self->addPhostingTar( 'anon_ftpstat', $domainId, "anon-ftpstat", "directory" => $path );
  $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_anon_ftpstat ) if $cid_anon_ftpstat;
}

sub setDomainPhostingConfContent {
  my ( $self, $domainId, $domainName, $path, $includePtr ) = @_;

  my @include = @{$includePtr};

  my $root = $self->{domainNodes}->{$domainId};

  my @phostings = $root->getChildren('phosting');
  if ( scalar(@phostings) < 1 ) {
    printToError('Error: setDomainPhostingConfContent: there are no node "phosting"');
    return;
  }

  if ( $#include != -1 ) {
    my $cid_conf = $self->addPhostingTar( 'conf', $domainId,
      "conf",
      "directory" => $path,
      "include"   => @include
    );
    $phostings[0]->getChild( 'content', 1, 1 )->addChild( $cid_conf ) if $cid_conf;
  }
}

sub setDomainFrontPageAdmin {
  my ( $self, $domainId, $fpAdm, $fpPass, $fpPassType) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my $phosting = $root->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainFrontPageAdmin: there are no node "phosting"');
    return;
  }

  my $item =
    XmlNode->new( 'fpuser',
    'attributes' => { 'name' => $fpAdm } );
  if ( $fpPass ) {
    my $passNode = &makePasswordNode( $fpPass, $fpPassType );
    $item->addChild($passNode);
  }

  $phosting->getChild( 'preferences', 1 )->addChild($item);
}

sub setDomainLogrotation {
  my ( $self, $domainId, $logRotationPtr) = @_;

  my %logRotation = %{$logRotationPtr};

  my $root = $self->{domainNodes}->{$domainId};

  my $phosting = $root->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainLogrotation: there are no node "phosting"');
    return;
  }

  my $node;
  if (%logRotation)
  {
    $self->makeLogrotationNode( $phosting->getChild( 'preferences', 1 ), \%logRotation );
  }
}

sub setDomainAnonFtp {
  my ( $self, $domainId, $domainName, $propsPtr, $pub_path, $incoming_path ) = @_;

  my %props = %{$propsPtr};

  my $root = $self->{domainNodes}->{$domainId};

  my $phosting = $root->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainAnonFtp: there are no node "phosting"');
    return;
  }

  my $anonRoot = XmlNode->new('anonftp');
  if ( $props{'status'} =~ /true/ ) {
    $anonRoot->setAttribute( 'pub', 'true' );
  }
  if ( $props{'incoming'} =~ /true/ ) {
    $anonRoot->setAttribute( 'incoming', 'true' );
  }
  if ( defined( $props{'max_conn'} ) ) {
    makeAnonftpLimitNode( $anonRoot, 'max-connections',
      $props{'max_conn'} );
  }
  if ( defined( $props{'bandwidth'} ) ) {
    makeAnonftpLimitNode( $anonRoot, 'bandwidth',
      $props{'bandwidth'} );
  }
  if ( defined( $props{'quota'} ) ) {
    makeAnonftpLimitNode( $anonRoot, 'incoming-disk-quota',
      $props{'quota'} );
  }
  if ( defined( $props{'display_login'} ) ) {
    $anonRoot->setAttribute( 'display-login', $props{'display_login'} );
  }
  if ( $props{'incoming_readable'} =~ /true/ ) {
    makeAnonftpPermissionNode( $anonRoot, 'incoming-download' );
  }
  if ( $props{'incoming_subdirs'} =~ /true/ ) {
    makeAnonftpPermissionNode( $anonRoot, 'incoming-mkdir' );
  }
  if ( defined( $props{'login_text'} ) ) {
    my $loginMessageNode =
      XmlNode->new( 'login-message',
      'content' => $props{'login_text'} );
    $anonRoot->addChild($loginMessageNode);
  }
  if ( defined($pub_path) && (-d $pub_path) ) {
    my $cid = $self->addPhostingTar( 'ftp_pub', $domainId, "anonftp.pub",  "directory" => $pub_path );
    $anonRoot->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
  }
  if ( defined($incoming_path) && (-d $incoming_path) ) {
    my $cid_incoming = $self->addPhostingTar( 'ftp_incoming', $domainId, "anonftp.incoming", "directory" => $incoming_path );
    $anonRoot->getChild( 'content', 1, 1 )->addChild( $cid_incoming ) if $cid_incoming;
  }

  $phosting->getChild( 'preferences', 1 )->addChild($anonRoot);
}

sub addDomainProtectedDir {
  my ( $self, $domainId, $pdirPath, $pdirTitle, $pdirNonSSL, $pdirSSL, $pdirCGI, $pdirUsersPtr) = @_;

  my @pdirUsers = @{$pdirUsersPtr};

  my $parent = $self->{domainNodes}->{$domainId};

  my $phosting = $parent->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: addDomainProtectedDir: there are no node "phosting"');
    return;
  }

  my $root = XmlNode->new('pdir');

  # workaround of CLI inabliity to create '' directory.
  if ( $pdirPath eq '' ) {
    $pdirPath = '/';
  }

  $root->setAttribute( 'name', $pdirPath );
  if ($pdirTitle) {
    $root->setAttribute( 'title', $pdirTitle );
  }
  if ($pdirNonSSL) {
    $root->setAttribute( 'nonssl', $pdirNonSSL );
  }
  if ($pdirSSL) {
    $root->setAttribute( 'ssl', $pdirSSL );
  }
  if ($pdirCGI) {
    $root->setAttribute( 'cgi', $pdirCGI );
  }

  my $userNode;
  my $item;
  for my $user (@pdirUsers) {
    $userNode = XmlNode->new('pduser');
    $root->addChild($userNode);
    $userNode->setAttribute( 'name', $user->{'login'} );

    $item =
      makePasswordNode( $user->{'passwd'},
                        normalizePasswordType( $user->{'passwdType'} ) );
    $userNode->addChild($item);

    $root->addChild($userNode);
  }

  $phosting->getChild( 'preferences', 1 )->addChild($root);
}

sub addDomainWebUser {
  my ( $self, $domainId, $domainName, $sysuserPtr, $scriptingPtr, $webUserHome, $privatePath ) = @_;

  my %sysuser = %{$sysuserPtr};
  my %scripting = %{$scriptingPtr};

  my $parent = $self->{domainNodes}->{$domainId};

  my $phosting = $parent->getChild('phosting');
  if ( not $phosting  ) {
    printToError('Error: addDomainWebUser: there are no node "phosting"');
    return;
  }

  my $root = XmlNode->new('webuser');

  $root->addChild($self->makeSysUser(\%sysuser));

  my $userName = $sysuser{'login'};
  $root->setAttribute( 'name', $userName );

  my $cid = $self->addPhostingTar( 'webuser_home', $domainId, "webuser_$userName", "directory" => $webUserHome );
  $root->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;

  my $cid_private = $self->addPhostingTar( 'webuser_private', $domainId,  "webuser_$userName.private", "directory" => $privatePath );
  $root->getChild( 'content', 1, 1 )->addChild( $cid_private ) if $cid_private;

  #
  # scripting
  #
  my $item = XmlNode->new('scripting');
  while ( my ( $xmlName, $fieldName ) = each(%webUserScripting) ) {
    if ( defined $scripting{$fieldName}
      and ( $scripting{$fieldName} eq 'true' or $scripting{$fieldName} eq 'false' ) )
    {
      $item->setAttribute( $xmlName, $scripting{$fieldName} );
    }
  }
  $root->addChild($item);

  #
  # end scripting
  #

  $phosting->getChild( 'webusers', 1 )->addChild($root);
}

sub setDomainWebalizer {
  my ( $self, $domainId, $directRef, $hiddenRefsPtr, $groupRefsPtr) = @_;

  my @hiddenRefs = @{$hiddenRefsPtr};
  my @groupRefs = @{$groupRefsPtr};

  my $parent = $self->{domainNodes}->{$domainId};

  my $phosting = $parent->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainWebalizer: there are no node "phosting"');
    return;
  }

  my $node = XmlNode->new("webalizer");

  if ($directRef)
  {
    $node->setAttribute( 'hide-direct-referrer', 'true' );
  }

  for my $ref (@hiddenRefs) {
    $node->addChild(
      XmlNode->new(
        "webalizer-hidden-referrer", 'content' => $ref
      )
    );
  }

  for my $groupRef (@groupRefs ) {
    $node->addChild(
      XmlNode->new(
        'webalizer-group-referrer',
        'content'    => $groupRef->{'ref'},
        'attributes' => { 'group-name' => $groupRef->{'name'} }
      )
    );
  }

  $phosting->getChild( 'preferences', 1 )->addChild($node);
}

sub setDomainSitebuilder {
  my ( $self, $domainId, $sbSitesPtr, $enabled) = @_;

  my @sbSites = @{$sbSitesPtr};

  my $parent = $self->{domainNodes}->{$domainId};

  my $phosting = $parent->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainWebalizer: there are no node "phosting"');
    return;
  }

  my $sbNode = XmlNode->new('sb-domain');

  if (defined($enabled)) {
  	$sbNode->setAttribute('enabled', $enabled);
  }

  for my $ptrHash (@sbSites ) {
    $sbNode->addChild(
      XmlNode->new( 'sb-site-id', 'content' => $ptrHash->{'sb_site_id'} )
    );
    $sbNode->addChild(
      XmlNode->new(
        'sb-siteowner-id', 'content' => $ptrHash->{'sb_siteowner_id'}
      )
    );
    $sbNode->addChild(
      XmlNode->new(
        'sb-siteowner-login',
        'content' => $ptrHash->{'sb_siteowner_login'}
      )
    );
  }

  $phosting->getChild( 'preferences', 1 )->addChild( $sbNode );
}

sub setDomainPerfomance {
  my ( $self, $domainId, $max_connection, $traffic_bandwidth) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if (!defined($parent) ) {
    printToError('Error: setDomainPerfomance: empty parent node');
    return;
  }

  my $phosting = $parent->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainPerfomance: there are no node "phosting"');
    return;
  }

  my $node = XmlNode->new('perfomance');
  $node->addChild(
    XmlNode->new(
      'max-connections', 'content' => $max_connection
    )
  );
  $node->addChild(
    XmlNode->new(
      'traffic-bandwidth', 'content' => $traffic_bandwidth
    )
  );

  $phosting->getChild( 'preferences', 1 )->addChild( $node );
}

sub setDomainWebStat{
  my ( $self, $domainId, $stat_ttl) = @_;

  my $parent = $self->{domainNodes}->{$domainId};
  if (!defined($parent) ) {
    printToError('Error: setDomainWebStat: empty parent node');
    return;
  }
  my $phosting = $parent->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: setDomainWebStat: there are no node "phosting"');
    return;
  }
  my $node = XmlNode->new('web-stat');
  $node->setAttribute( 'stat-ttl', $stat_ttl ) if $stat_ttl;
  $phosting->getChild( 'preferences', 1 )->addChild( $node );
}

sub setDomainShosting {
  my ( $self, $domainId, $forward) = @_;

  my $parent = $self->{domainNodes}->{$domainId};

  my $root = XmlNode->new( 'shosting', 'content' => $forward );

  $parent->addChild($root);
}

sub setDomainFhosting {
  my ( $self, $domainId, $forward) = @_;

  my $parent = $self->{domainNodes}->{$domainId};

  my $root = XmlNode->new( 'fhosting', 'content' => $forward );

  $parent->addChild($root);
}

### SubDomain functions ##

sub addSubDomain {
  my ( $self, $domainId, $subDomainId, $subDomainName, $subAsciiDomainName, $https, $shared_content, $sysuserPtr, $scriptingPtr) = @_;
  $subAsciiDomainName = $subDomainName if not $subAsciiDomainName;
  $self->regSubdomainObjectBackupPath( $self->getPhostingBackupPath( $domainId ), $subDomainId, $subAsciiDomainName, $domainId );

  my %sysuser = %{$sysuserPtr};
  my %scripting = %{$scriptingPtr};

  my $parent = $self->{domainNodes}->{$domainId};

  my $phosting = $parent->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: addSubdomain: there are no node "phosting"');
    return;
  }

  my $root = XmlNode->new('subdomain');
  $self->{subDomainNodes}->{$subDomainId} = $root;
  my $parentGuid = $parent->getGuid();
  $root->setAttribute( "guid", $self->getSubdomainGuid( $parentGuid, $subDomainId ) ) if $parentGuid;
  my $phostingGuid = $phosting->getGuid();
  $root->setAttribute( 'owner-guid', $phostingGuid ) if $phostingGuid;

  $root->setAttribute( 'name', $subDomainName );

  if ( 'true' eq $https ) {
    $root->setAttribute( 'https', 'true' );
  }
  if ( 'true' eq $shared_content ) {
    $root->setAttribute( 'shared-content', 'true' );
  }

  ## %sysuser ##

  if (%sysuser)
  {
    $root->addChild($self->makeSysUser(\%sysuser));
  }

  ## %scripting ##

  my $scripringItem = XmlNode->new('scripting');
  while ( my ( $xmlName, $value ) = each(%scripting) ) {
      $scripringItem->setAttribute( $xmlName, $value );
  }
  $root->addChild($scripringItem);

  $phosting->getChild( 'subdomains', 1 )->addChild($root);

}

sub setSubDomainHttpdocsContent {
  my ( $self, $subDomainId, $subDomainName, $domainName, $path, $exclude_httpdocs_files_ptr ) = @_;

  my @exclude_httpdocs_files = @{$exclude_httpdocs_files_ptr};

  my $root = $self->{subDomainNodes}->{$subDomainId};

  my $cid_docroot = $self->addSubDomainPhostingTar( 'docroot', $subDomainId,
     "httpdocs",
    "directory" => $path,
    "exclude"   => \@exclude_httpdocs_files
  );
  $root->getChild( 'content', 1, 1 )->addChild( $cid_docroot ) if $cid_docroot;
}

sub setSubDomainHttpsdocsContent {
  my ( $self, $subDomainId, $subDomainName, $domainName, $path, $exclude_httpsdocs_files_ptr ) = @_;

  my @exclude_httpsdocs_files = @{$exclude_httpsdocs_files_ptr};

  my $root = $self->{subDomainNodes}->{$subDomainId};

  my $cid_docroot_ssl = $self->addSubDomainPhostingTar('docroot_ssl', $subDomainId,
    "httpsdocs",
    "directory" => $path,
    "exclude"   => \@exclude_httpsdocs_files
  );
  $root->getChild( 'content', 1, 1 )->addChild( $cid_docroot_ssl ) if $cid_docroot_ssl;
}

sub setSubDomainCgibinContent {
  my ( $self, $subDomainId, $subDomainName, $domainName, $path, $exclude_cgi_bin_files_ptr ) = @_;

  my @exclude_cgi_bin_files = @{$exclude_cgi_bin_files_ptr};

  my $root = $self->{subDomainNodes}->{$subDomainId};

  my $cid_cgi = $self->addSubDomainPhostingTar( 'cgi',  $subDomainId,
    "cgi-bin",
    "directory" => $path,
    "exclude"   => \@exclude_cgi_bin_files
  );
  $root->getChild( 'content', 1, 1 )->addChild( $cid_cgi ) if $cid_cgi;
}

sub setSubDomainConfContent {
  my ( $self, $subDomainId, $subDomainName, $domainName, $path ) = @_;

  my $root = $self->{subDomainNodes}->{$subDomainId};

  my $cid_conf = $self->addSubDomainPhostingTar( 'conf', $subDomainId,
    "conf",
    "directory" => $path
  );
  $root->getChild( 'content', 1, 1 )->addChild( $cid_conf ) if $cid_conf;
}

### Mail functions ##

sub addMail {
    my ( $self, $domainId, $mailId, $mailName, $passwd, $typePasswd ) = @_;

    $self->regMailnameObjectBackupPath( $self->getDomainsBackupPath( $domainId ), $mailId, $mailName, $domainId );

    my $parent = $self->{domainNodes}->{$domainId};

    my $mailService = $parent->getChild('mailsystem');
    if ( not $mailService ) {
      printToError("Error: Packer.addMail: there are no node 'mailsystem'");
      return;
    }

    my $root = XmlNode->new( 'mailuser', 'attributes' => {
        'id'   => $mailId,
        'name' => $mailName
        } );
    my $parentGuid = $parent->getGuid();
    $root->setAttribute( 'guid', $self->getMailnameGuid( $parentGuid, $mailId ) ) if $parentGuid;
    $root->setAttribute( 'owner-guid', $parentGuid ) if $parentGuid;

    my $props = $root->getChild( 'properties', 1 );
    $root->getChild( 'limits-and-permissions', 1 );
    $root->getChild( 'preferences', 1 );

    my $item;
    if ( defined($passwd) && $passwd ) {
      if ( $passwd =~ /NULL/ ) {
        $item = makePasswordNode( '', 'plain' );
      }
      else {
        $item = makePasswordNode( $passwd, $typePasswd );
      }

      $props->addChild( $item );
    }

    $self->{mailNodes}->{$mailId} = $root;

    $mailService->getChild( 'mailusers', 1 )->addChild($root);

    $self->{mailnamesNodes}->{$mailId} = $root;
}

sub setMailMaxButtonLength {
    my ( $self, $mailId, $length ) = @_;

    my $root = $self->{mailNodes}->{$mailId};

    $root->setAttribute( 'max-button-length', $length );
}

sub setMailSkin {
    my ( $self, $mailId, $skin ) = @_;

    my $root = $self->{mailNodes}->{$mailId};

    $root->setAttribute( 'skin', $skin );
}

sub setMailLockScreen {
    my ( $self, $mailId ) = @_;

    my $root = $self->{mailNodes}->{$mailId};

    $root->setAttribute( 'lock-screen', 'false' );
}

sub addMailPermission {
    my ( $self, $mailId, $name, $value ) = @_;

    my $root = $self->{mailNodes}->{$mailId};

    $self->makeMailuserPermissionNode( $root, $name, $value );
}

sub setMailBoxQuota {
    my ( $self, $mailId, $mbox_quota ) = @_;

    my $root = $self->{mailNodes}->{$mailId};

    $root->setAttribute( 'mailbox-quota', $mbox_quota );
}

sub setMailBox {
    my ( $self, $mailId, $mailName, $domainAsciiName, $enabled, $path ) = @_;

    my $root = $self->{mailNodes}->{$mailId};
    if ( !defined($root) ) {
      printToError('Error: setMailBox: empty parent node');
      return;
    }

    my $item = XmlNode->new( 'mailbox', 'attributes' => { 'type' => 'mdir' } );
    $root->getChild( 'preferences', 1 )->addChild($item);

    $item->setAttribute( 'enabled', $enabled );

    my $cid = $self->addMailnameTar( 'mailbox', $mailId, "mdir",
      'directory' => $path );
    $item->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}

sub addMailAliase {
    my ( $self, $mailId, $name ) = @_;

    my $root = $self->{mailNodes}->{$mailId};
    if ( !defined($root) ) {
      printToError('Error: addMailAliase: empty parent node');
      return;
    }

    my $item = XmlNode->new( 'alias', 'content' => $name );
    $root->getChild( 'preferences', 1 )->addChild($item);
}

sub setMailRedirect {
    my ( $self, $mailId, $addr, $enabled ) = @_;

    my $root = $self->{mailNodes}->{$mailId};
    if ( !defined($root) ) {
      printToError('Error: setMailRedirect: empty parent node');
      return;
    }

    my $item = XmlNode->new( 'redirect',
      'content'    => $addr,
      'attributes' => { 'enabled' => ( $enabled eq 'true'? 'true': 'false') } );

    $root->getChild( 'preferences', 1 )->addChild($item);
}

sub setMailGroup {
    my ( $self, $mailId, $enabled, $membersPtr ) = @_;

    unless ( ref($membersPtr) =~ /ARRAY/ ) {
      printToError("Error: setMailGroup: membersPtr is not ref array");
      return;
    }

    my @members = @{$membersPtr};

    my $root = $self->{mailNodes}->{$mailId};
    if ( !defined($root) ) {
      printToError('Error: addMailRedirect: empty parent node');
      return;
    }


    $root->setAttribute( 'mailgroup-enabled', $enabled );

    my $item;
    foreach my $member (@members) {
      $item = XmlNode->new( 'mailgroup-member', 'content' => $member );
      $root->getChild( 'preferences', 1 )->addChild($item);
    }
}

sub setMailAutoresponders {
    my ( $self, $mailId, $mailName, $domainAsciiName, $dir, $enabled, $autorespondersPtr, $attachesPtr ) = @_;

    unless ( ref($autorespondersPtr) =~ /ARRAY/ ) {
      printToError("Error: setMailAutoresponders: autorespondersPtr is not ref array");
      return;
    }
    my @autoresponders = @{$autorespondersPtr};

    unless ( ref($attachesPtr) =~ /ARRAY/ ) {
      printToError("Error: setMailAutoresponders: attachesPtr is not ref array");
      return;
    }
    my @attaches = @{$attachesPtr};

    my $root = $self->{mailNodes}->{$mailId};
    if ( !defined($root) ) {
      printToError('Error: setMailAutoresponders: empty parent node');
      return;
    }

    my $autorespondersNode = XmlNode->new( 'autoresponders',
      'attributes' => { 'enabled' => $enabled } );

    my $cid = $self->addMailnameTar( 'attaches', $mailId,  "attach",
      "directory" => $dir );
    $autorespondersNode->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;

    foreach my $file (@attaches) {
      $autorespondersNode->addChild(
        XmlNode->new( 'attach', 'attributes' => { 'file' => $file } ) );
    }

    foreach my $autoresponder (@autoresponders) {
      my $node = $self->makeAutoresponderNode ($autoresponder, $mailName);
      $autorespondersNode->addChild($node);
    }

    $root->getChild( 'preferences', 1 )->addChild($autorespondersNode);
}

sub setMailAddressbook {
    my ( $self, $mailId, $paramsPtr ) = @_;

    my @params = @{$paramsPtr};

    my $parent = $self->{mailNodes}->{$mailId};
    if ( !defined($parent) ) {
      printToError('Error: setMailAddressbook: empty parent node');
      return;
    }

    my $node = XmlNode->new('addressbook');

    my %objectFields = (
      'id'           => 'id',
      'name'         => 'name',
      'alias'        => 'alias',
      'email'        => 'email',
      'title'        => 'title',
      'company'      => 'company',
      'home-phone'   => 'homephone',
      'work-phone'   => 'workphone',
      'mobile-phone' => 'cellphone',
      'fax'          => 'fax',
      'home-address' => 'homeaddress',
      'work-address' => 'workaddress',
      'notes'        => 'notes'
      );

    foreach my $ptrHash (@params) {
      if ( $ptrHash->{'object_type'} eq 'Group' ) {

        my $group =
          XmlNode->new( 'addressbook-group',
          'attributes' => { 'id' => $ptrHash->{'object_id'} } );

        my $members = $ptrHash->{'object_members'};
        while ( $members =~ /"(.*?)"/g ) {
          my $member = $1;
          chomp $member;

          $group->addChild(
            XmlNode->new(
              'addressbook-member', 'attributes' => { 'id' => $member }
            )
          );
        }

        $node->addChild($group);
      }
      else {
        my $user = XmlNode->new('addressbook-contact');
        my ( $field, $attr );
        while ( ( $field, $attr ) = each %objectFields ) {
          if ( defined $ptrHash->{ 'object_' . $attr } ) {

            # Here 'encoded' attribute means there is no meaning to
            # convert data to UTF as it was already converted manually
            my $value = $ptrHash->{ 'object_' . $attr };
            $value = $self->{base64}->{'ENCODE'}->($value) if $attr ne 'id';
            $user->setAttribute( $field, $value );
          }
        }
        $node->addChild($user);
      }
    }

    $parent->getChild( 'preferences', 1 )->addChild($node);
}

sub setMailSpamSettings {
    my ( $self, $mailId, $mailName, $domainAsciiName, $status, $server_conf,
      $action, $requiredScore, $modify_subj, $subj_text,
      $blackListPtr, $whiteListPtr, $unblackListPtr, $unwhiteListPtr, $path, $server ) = @_;

    my @blackList = @{$blackListPtr};
    my @whiteList = @{$whiteListPtr};
    my @unblackList = @{$unblackListPtr};
    my @unwhiteList = @{$unwhiteListPtr};

    my $parent;
    if (defined($mailId)) {
    	$parent = $self->{mailNodes}->{$mailId};
    }elsif (defined($server)) {
    	$parent = $self->{serverNode};
    }
    if ( !defined($parent) ) {
      printToError('Error: setMailSpamSettings: empty parent node');
      return;
    }

    my $item = XmlNode->new( 'spamassassin' );

    $item->setAttribute( 'status', $status ) if defined($status);
    $item->setAttribute( 'action', $action ) if ( defined($action) && $action );
    $item->setAttribute( 'server-conf', $server_conf )
      if ( defined($server_conf) && $server_conf );
    $item->setAttribute( 'hits', $requiredScore )
      if ( defined($requiredScore) && $requiredScore );
    $item->setAttribute( 'modify-subj', $modify_subj )
      if ( defined($modify_subj) && $modify_subj );
    $item->setAttribute( 'subj-text', $subj_text )
      if ( defined($subj_text) && $subj_text );
    $item->setAttribute( 'max-spam-threads', $server->{'spamfilter_max_children'})
      if ( defined($server->{'spamfilter_max_children'}) && $server->{'spamfilter_max_children'});

    foreach my $blMail (@blackList) {
      $item->addChild( XmlNode->new( 'blacklist-member', 'content' => $blMail ) );
    }
    foreach my $wtMail (@whiteList) {
      $item->addChild( XmlNode->new( 'whitelist-member', 'content' => $wtMail ) );
    }
    foreach my $ublMail (@unblackList) {
      $item->addChild(
        XmlNode->new( 'unblacklist-member', 'content' => $ublMail ) );
    }
    foreach my $uwtMail (@unwhiteList) {
      $item->addChild(
        XmlNode->new( 'unwhitelist-member', 'content' => $uwtMail ) );
    }

    if (defined($mailId)) {
      my $cid = $self->addMailnameTar( 'spam-assasin', $mailId, "sa", "directory" => $path );
      $item->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
    }

    if (defined($mailId)) {
	    $parent->getChild( 'preferences', 1 )->addChild($item);
    }elsif(defined($server)) {
    	    $parent->addChild($item);
    }
}

sub setMailVirusSettings {
    my ( $self, $mailId, $state ) = @_;

    my $parent = $self->{mailNodes}->{$mailId};
    if ( !defined($parent) ) {
      printToError('Error: setMailVirusSettings: empty parent node');
      return;
    }

    my $item = XmlNode->new('virusfilter');
    $item->setAttribute( 'state', $state );

    $parent->getChild( 'preferences', 1 )->addChild($item);
}

sub addMailCustomButton {
    my ( $self, $mailId, $id, $optionsPtr, $customButtonsDir, $icon ) = @_;

    my $parent = $self->{mailNodes}->{$mailId};
    if ( !defined($parent) ) {
      printToError('Error: addMailCustomButton: empty parent node');
      return;
    }

    my $node = $self->makeCustomButtonNode( 'mails', $mailId, $id, $optionsPtr, $customButtonsDir, $icon );

    $parent->getChild( 'preferences', 1 )->addChild($node);
}

####### Siteapps functions #######

sub addDomainSapp {
  my ( $self, $domainId, $sappId, $name, $version, $release, $description,
          $isCommercial, $isIntegrated, $prefix, $isSsl, $licenseType ) = @_;

  my $root = $self->{domainNodes}->{$domainId};

  my $phosting = $root->getChild('phosting');
  if ( not $phosting ) {
    printToError('Error: addDomainSiteApp: there are no node "phosting"');
    return;
  }

  my $sapp_installed = $self->makeSappNode($sappId, $name, $version, $release, $description,
            $isCommercial, $isIntegrated, $prefix, $isSsl, $licenseType );

  $self->{sappNodes}->{$sappId} = $sapp_installed;
  $phosting->getChild( 'applications', 1 )->addChild($sapp_installed);
}

sub addSubDomainSapp {
  my ( $self, $subDomainId, $sappId, $name, $version, $release, $description,
            $isCommercial, $isIntegrated, $prefix, $isSsl, $licenseType ) = @_;

  my $root = $self->{subDomainNodes}->{$subDomainId};

  my $sapp_installed = $self->makeSappNode($sappId, $name, $version, $release, $description,
            $isCommercial, $isIntegrated, $prefix, $isSsl, $licenseType);

  $self->{sappNodes}->{$sappId} = $sapp_installed;
  $root->getChild( 'applications', 1 )->addChild($sapp_installed);
}

sub setSappParams {
  my ( $self, $sappId, $paramsPtr ) = @_;

  my %params = %{$paramsPtr};

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  while ( my ( $param_name, $param_value ) = each %params )
  {
    my $name  = $self->{base64}->{'ENCODE'}->($param_name);
    my $value = $self->{base64}->{'ENCODE'}->($param_value);

    $sapp_installed->addChild(
      XmlNode->new(
        'sapp-param',
        'children' => [
          XmlNode->new(
            'sapp-param-name',
            'attributes' => { 'encoding' => 'base64' },
            'content'    => $name
          ),
          XmlNode->new(
            'sapp-param-value',
            'attributes' => { 'encoding' => 'base64' },
            'content'    => $value
          )
        ]
      )
    );
  }
}

sub setDomainSappScriptsContent {
  my ( $self, $domainId, $sappId, $scriptsPath, $httpsScriptsPath ) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  my $dump_id   = $sappId;

  my $cid = $self->addPhostingTar( 'sapp-scripts', $domainId, "sapp-scripts." . $dump_id, "directory" => $scriptsPath );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
  $cid  = $self->addPhostingTar( 'sapp-https-scripts', $domainId, "sapp-https-scripts." . $dump_id, "directory" => $httpsScriptsPath );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}

sub setSubDomainSappScriptsContent {
  my ( $self, $subdomainId, $sappId, $scriptsPath, $httpsScriptsPath) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  my $dump_id   = $sappId;

  my $cid = $self->addSubDomainPhostingTar( 'sapp-scripts', $subdomainId, "sapp-scripts." . $dump_id, "directory" => $scriptsPath );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
  $cid  = $self->addSubDomainPhostingTar( 'sapp-https-scripts', $subdomainId, "sapp-https-scripts." . $dump_id, "directory" => $httpsScriptsPath );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}

sub setDomainSappCgibinContent {
  my ( $self, $domainId, $sappId, $path, $include ) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  my $cid        = $self->addPhostingTar( 'sapp-cgi-bin', $domainId,
    "sapp-cgi-bin." . $sappId,
    "directory"    => $path,
    "include"      => $include,
    "no_recursion" => 1
  );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}

sub setSubDomainSappCgibinContent {
  my ( $self, $subdomainId, $sappId, $path, $include ) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  my $cid        = $self->addSubDomainPhostingTar( 'sapp-cgi-bin', $subdomainId,
    "sapp-cgi-bin." . $sappId,
    "directory"    => $path,
    "include"      => $include,
    "no_recursion" => 1
  );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}


sub setDomainSappHttpContent {
  my ( $self, $domainId, $sappId, $path, $include ) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  my $cid = $self->addPhostingTar( 'sapp-htdocs', $domainId,
    "sapp-htdocs." . $sappId,
    "directory"    => $path,
    "include"      => $include,
    "no-recursion" => 1
  );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}

sub setSubDomainSappHttpContent {
  my ( $self, $subdomainId, $sappId, $path, $include ) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  my $cid = $self->addSubDomainPhostingTar( 'sapp-htdocs', $subdomainId,
    "sapp-htdocs." . $sappId,
    "directory"    => $path,
    "include"      => $include,
    "no-recursion" => 1
  );
  $sapp_installed->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
}

sub addDomainSappDatabase {
  my ( $self, $dbId, $dbServerId, $domainId, $sappId, $dbName, $dbType, $optionalPtr, $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr ) = @_;

  $self->regDatabaseObjectBackupPath( $self->getDomainsBackupPath( $domainId ), $dbId, $dbName, $dbServerId );
  my $parent = $self->{sappNodes}->{$sappId};
  my $domain = $self->{domainNodes}->{$domainId};

  my $node = $self->makeDatabaseNode( $dbId, $dbName, $dbType, 'domain', $domainId, $optionalPtr,
                                      $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr );
  my $domainGuid = $domain->getGuid();
  $node->setAttribute( 'guid', $self->getDatabaseGuid( $domainGuid, $dbId ) ) if $domainGuid;
  $node->setAttribute( 'owner-guid', $domainGuid ) if $domainGuid;

  $parent->addChild($node);
}

sub addSubDomainSappDatabase {
  my ( $self, $dbId, $dbServerId, $subdomainId, $sappId, $dbName, $dbType, $optionalPtr, $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr ) = @_;

  $self->regDatabaseObjectBackupPath( $self->getSubdomainsBackupPath( $subdomainId ), $dbId, $dbName, $dbServerId );
  my $subdomain = $self->{subDomainNodes}->{$subdomainId};

  my $parent = $self->{sappNodes}->{$sappId};

  my $node = $self->makeDatabaseNode( $dbId, $dbName, $dbType, 'subdomain', $subdomainId, $optionalPtr,
                                      $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr );
  my $subDomainGuid = $subdomain->getGuid();
  $node->setAttribute( 'guid', $self->getDatabaseGuid( $subDomainGuid, $dbId ) ) if $subDomainGuid;
  $node->setAttribute( 'owner-guid', $subDomainGuid ) if $subDomainGuid;

  $parent->addChild($node);
}

sub addDomainSappCustomButton {
    my $self = shift;
    $self->addSappCustomButton( 'domains', @_ );
}

sub addSubDomainSappCustomButton {
    my $self = shift;
    $self->addSappCustomButton( 'subdom', @_ );
}


sub addSappCustomButton {
    my ( $self, $parentType, $parentid, $sappId, $id, $optionsPtr, $customButtonsDir, $icon ) = @_;

    my $parent = $self->{sappNodes}->{$sappId};
    if ( !defined($parent) ) {
      printToError('Error: addSappCustomButton: empty parent node');
      return;
    }

    my $node = $self->makeCustomButtonNode( $parentType, $parentid, $id, $optionsPtr, $customButtonsDir, $icon );

    $parent->addChild($node);
}

####### Private members #######

sub normalizeCountry {
    my ($country) = @_;
    return 'RU' if $country eq 'SU';
    return 'GB' if $country eq 'UK';
    return $country;
}

sub normalizePasswordType {
    my ($type) = @_;
    return "encrypted" if $type eq "crypt";
    return $type;
}

sub insertLimitNode {
    my ( $self, $node, $name, $value ) = @_;
    my $limits = $node->getChild( 'limits-and-permissions', 1 );
    $limits->addChild( XmlNode->new( "limit",
        "attributes" => { "name" => $name },
        "content"    => $value ) );
}

sub makePermissionNode {
  my ($self, $parent, $name, $allowed) = @_;

  #[Bug 111477] Value "-1" for attribute allowed of permission is not among the enumerated set
  if ($allowed =~ /-1/) {
   	$allowed = 'false';
  }
  $parent->addChild(  XmlNode->new('permission', 'attributes' => { 'name' => $name, 'value' => $allowed}) );
}

sub makeMailuserPermissionNode {
   my ($sefl, $parent, $name, $allowed) = @_;
  my $permissions = $parent->getChild( 'limits-and-permissions', 1 );

  $permissions->addChild(XmlNode->new('mailuser-permission', 'attributes' => { 'name' => $name, 'value' => $allowed}));
}

#####################################################
#
#             Backup content structure
#
#####################################################


sub getPhostingGuid{
   my ( $self, $ownerguid, $id ) = @_;
   return "$ownerguid\_ph_$id";
}

sub getDatabaseGuid{
   my ( $self, $ownerguid, $id ) = @_;
   return "$ownerguid\_db_$id";
}

sub getSubdomainGuid{
   my ( $self, $ownerguid, $id ) = @_;
   return "$ownerguid\_sb_$id";
}

sub getMailnameGuid{
   my ( $self, $ownerguid, $id ) = @_;
   return "$ownerguid\_mn_$id";
}

sub regObjectBackupPath{
 my( $self, $basePath, $group, $key, $objectId, $parentId ) = @_;
 $basePath .= '/' if $basePath;
 my $path = "$basePath$group";
 $path = $basePath if $group eq 'admin';
 $self->{backupPath}->{"$group\_$key"} = $path;
 $self->{objectId}->{"$group\_$key"} = $objectId;
 $self->{parentId}->{"$group\_$key"} = $parentId;
 return $self->getBackupPath( $group, $key, undef, 0 );
}

sub regAdminObjectBackupPath{
 my( $self, $basePath ) = @_;
 return $self->regObjectBackupPath( $basePath, 'admin', 'admin', '' );
}

sub regResellersObjectBackupPath{
 my( $self, $basePath, $key, $resellerLogin ) = @_;
 return $self->regObjectBackupPath( $basePath, 'resellers', $key, $self->{fnamecreator}->normalize_short_string( $resellerLogin, $key ) );
}

sub regClientObjectBackupPath{
 my( $self, $basePath, $key, $clientLogin ) = @_;
 return $self->regObjectBackupPath( $basePath, 'clients', $key, $self->{fnamecreator}->normalize_short_string( $clientLogin, $key ) );
}

sub regDomainObjectBackupPath{
 my( $self, $basePath, $key, $domainAscName ) = @_;
 return $self->regObjectBackupPath( $basePath, 'domains', $key, $self->{fnamecreator}->normalize_long_string( $domainAscName, $key ) );
}

sub regMailnameObjectBackupPath{
 my( $self, $basePath, $key, $mailName, $domainId ) = @_;
 return $self->regObjectBackupPath( $basePath, 'mailnames', $key, $self->{fnamecreator}->normalize_short_string( $mailName, $key ), $domainId );
}

sub regDatabaseObjectBackupPath{
 my( $self, $basePath, $key, $databaseName, $dbServerId ) = @_;
 return $self->regObjectBackupPath( $basePath, 'databases', $key, $self->{fnamecreator}->normalize_short_string( "$databaseName\_$dbServerId", $key ) );
}

sub regSubdomainObjectBackupPath{
 my( $self, $basePath, $key, $subdomainName, $domainId ) = @_;
 return $self->regObjectBackupPath( $basePath, 'subdomains', $key, $self->{fnamecreator}->normalize_short_string( $subdomainName, $key ), $domainId );
}

sub regPhostingObjectBackupPath{
 my( $self, $basePath, $key ) = @_;
 return $self->regObjectBackupPath( $basePath, 'phosting', $key, '' );
}

sub getObjectId{
 my( $self, $group, $key ) = @_;
 $key = "$group\_$key";
 return 'phosting' if $group eq 'phosting';
 return 'admin' if $group eq 'admin';
 return $self->{objectId}->{$key};
}

sub getParentId{
 my( $self, $group, $key ) = @_;
 $key = "$group\_$key";
 return $self->{parentId}->{$key}  if exists $self->{parentId} && exists $self->{parentId}->{$key};
 return undef;
}

sub getPhostingParentId{
 my( $self, $key ) = @_;
 return $key;
}

sub getSubdomainParentId{
 my( $self, $key ) = @_;
 return $self->getParentId( "subdomains", $key );
}

sub getMailNameParentId{
 my( $self, $key ) = @_;
 return $self->getParentId( "mailnames", $key );
}


sub getBackupPath{
 my( $self, $group, $key, $fileName, $onlyFileName ) = @_;
 $key = "$group\_$key";
 my $res = '';
 if( not $onlyFileName ) {
     $res .=  $self->{backupPath}->{$key};
     $res .= '/' . $self->{objectId}->{$key} if $group ne 'phosting' and $group ne 'admin';
 }
 $res = $self->{fnamecreator}->getFileName( $res, $self->{backupname}, $self->{objectId}->{$key}, $fileName ) if defined $fileName;
 return $res;
}

sub getAdminBackupPath{
 my( $self, $fileName, $onlyFileName ) = @_;
 return $self->getBackupPath( 'admin', 'admin', $fileName, $onlyFileName );
}

sub getResellersBackupPath{
 my( $self, $key, $fileName, $onlyFileName ) = @_;
 return $self->getBackupPath( 'resellers', $key, $fileName, $onlyFileName );
}

sub getClientsBackupPath{
 my $self = shift;
 my $key = shift;
 return $self->getBackupPath( 'clients', $key, @_ ) if exists $self->{objectId}->{"clients_$key"};
 return $self->getResellersBackupPath( $key, @_ );
}

sub getDomainsBackupPath{
 my $self = shift;
 return $self->getBackupPath( 'domains', @_ );
}

sub getPhostingBackupPath{
 my $self = shift;
 return $self->getBackupPath( 'phosting', @_ );
}

sub getDatabasesBackupPath{
 my $self = shift;
 return $self->getBackupPath( 'databases', @_ );
}

sub getMailnamesBackupPath{
 my $self = shift;
 return $self->getBackupPath( 'mailnames', @_ );
}

sub getSubdomainsBackupPath{
 my $self = shift;
 return $self->getBackupPath( 'subdomains', @_ );
}

sub getAdminObjectId{
 my( $self ) = @_;
 return $self->getObjectId( 'admin', 'admin' );
}

sub getResellerObjectId{
 my( $self, $key ) = @_;
 return $self->getObjectId( 'resellers', $key );
}

sub getClientObjectId{
 my( $self, $key ) = @_;
 return $self->getResellerObjectId( $key ) if exists $self->{objectId}->{"resellers_$key"};
 return $self->getObjectId( 'clients', $key );
}

sub getDomainObjectId{
 my( $self, $key ) = @_;
 return $self->getObjectId( 'domains', $key );
}

sub getPhostingObjectId{
 my( $self, $key ) = @_;
 return $self->getObjectId( 'phosting', $key );
}

sub getDatabaseObjectId{
 my( $self, $key ) = @_;
 return $self->getObjectId( 'databases', $key );
}

sub getMailnameObjectId{
 my( $self, $key ) = @_;
 return $self->getObjectId( 'mailnames', $key );
}

sub getSubdomainObjectId{
 my( $self, $key ) = @_;
 return $self->getObjectId( 'subdomains', $key );
}

sub addAdminTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    return $self->addTar( $cid_type, $self->getAdminBackupPath( $file ) , @_);
}

sub addResellerTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    return $self->addTar( $cid_type, $self->getResellersBackupPath($id, $file ), @_);
}

sub addClientTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    return $self->addResellerTar( $cid_type, $id, $file, @_ ) if exists $self->{resellersNodes}->{$id};
    return $self->addTar( $cid_type, $self->getClientsBackupPath($id, $file ), @_);
}

sub addDomainTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    return $self->addTar( $cid_type, $self->getDomainsBackupPath($id, $file ), @_);
}

sub addDomainMailListTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    my $referer = $self->getDomainMlTarReferer( $id, @_ );
    if( $referer  ) {
      return $self->addRefererCid( $referer , $cid_type);
    }
    else{
      return $self->addTar( $cid_type, $self->getDomainsBackupPath($id, $file ), @_);
    }
}

sub addPhostingTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    my $referer = $self->getDomainPhTarReferer( $self->getPhostingParentId( $id ), @_ );
    if( $referer  ) {
      return $self->addRefererCid( $referer , $cid_type);
    }
    else{
      return $self->addTar( $cid_type, $self->getPhostingBackupPath($id, $file ), @_);
    }
}

sub addSubDomainTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    return $self->addTar( $cid_type, $self->getSubdomainsBackupPath($id, $file ), @_);
}

sub getTarReferer{
  my( $self, $referrer, %options ) = @_;
  if( $referrer && exists $options{'directory'} && index( $options{'directory'}, $referrer->[1] )==0 ){
      my @res;
      push @res, $referrer->[0];
      my $ref = substr( $options{'directory'}, length($referrer->[1]) );
      while( index( $ref, '/' )==0 ){ $ref = substr( $ref, 1 ); }
      push @res, $ref;
      push @res, $options{'directory'};
      return \@res;
    }
  return undef;
}

sub getDomainMlTarReferer{
  my $self = shift;
  my $domainId = shift;
  return $self->getTarReferer( \@{$self->{domainMailListContent}->{$domainId}}, @_ ) if exists $self->{domainMailListContent}->{$domainId};
  return undef;
}

sub getDomainPhTarReferer{
  my $self = shift;
  my $domainId = shift;
  return $self->getTarReferer( \@{$self->{domainVhost}->{$domainId}}, @_ ) if exists $self->{domainVhost}->{$domainId};
  return undef;
}

sub addSubDomainPhostingTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    my $referer = $self->getDomainPhTarReferer( $self->getSubdomainParentId( $id ), @_ );
    if( $referer ) {
      return $self->addRefererCid( $referer , $cid_type );
    }
    else {
      return $self->addTar( $cid_type, $self->getSubdomainsBackupPath($id, $file ), @_);
    }
}

sub getDomainMailTarReferer{
  my $self = shift;
  my $domainId = shift;
  return $self->getTarReferer( \@{$self->{domainMailContent}->{$domainId}}, @_ ) if exists $self->{domainMailContent}->{$domainId};
  return undef;
}

sub addMailnameTar{
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    my $file = shift;
    my $referrer = $self->getDomainMailTarReferer( $self->getMailNameParentId( $id ), @_ );
    my $domainId = $self->getMailNameParentId( $id );
    if( $referrer ){
      return $self->addRefererCid( $referrer, $cid_type );
    }
    else {
      return $self->addTar( $cid_type, $self->getMailnamesBackupPath($id, $file ), @_);
    }
}

sub addTar {
    my $self = shift;
    my $cid_type = shift;
    my $dstFile = shift;
    return undef if exists $self->{skip_content};
    my $id = $self->{storage}->addTar( $dstFile, @_);
    return $self->makeCidNode( $id, $cid_type ) if $id;
    return undef;
}

sub addDbTar {
    my $self = shift;
    my $cid_type = shift;
    my $id = shift;
    return $self->addDb( $cid_type, $self->getDatabasesBackupPath( $id, '' ), @_);
}

sub addSubdomainDbTar {
    my $self = shift;
    my $cid_type = shift;
    my $subdomainId = shift;
    my $id = shift;
    return $self->addDb( $cid_type, $self->getSubdomainsBackupPath( $subdomainId, "db$id" ), @_);
}

sub addDb {
    my $self = shift;
    my $cid_type = shift;
    my $dstFile = shift;
    return undef if exists $self->{skip_content};
    my $id = $self->{storage}->addDb( $dstFile, @_);
    return $self->makeCidNode( $id, $cid_type ) if $id;
    return undef;
}

sub makeCidNode{
   my ($self, $id, $cid_type ) = @_;
   my $cid = XmlNode->new( 'cid' );
   $cid->setAttribute( 'type', $cid_type );
   $cid->setAttribute( 'unpacksize', $self->{storage}->getFilesUnpackSizeFromId($id) );
   $cid->setAttribute( 'path', $self->{storage}->getFilePathFromId( $id ) );
   my $files = $self->{storage}->getFilesFromId( $id );
   foreach my $filedata( @{$files} ){
     $cid->addChild( XmlNode->new( 'content-file', 'content' => $filedata->[0], 'attributes' => {'size' => $filedata->[1] } ) );
   }
   return $cid;
}

sub addRefererCid{
   my ($self, $referer, $cid_type ) = @_;
   if( $referer ){
     printToLog( "Add referrer cid for type '$cid_type', referer '$referer->[1]', from path '$referer->[2]'" );
     if( -e $referer->[2] ){
       my $cid = XmlNode->new( 'cid' );
       $cid->setAttribute( 'type', $cid_type );
       $cid->setAttribute( 'unpacksize', $referer->[0]->getAttribute( 'unpacksize' ) );
       $cid->setAttribute( 'path', $referer->[0]->getAttribute( 'path' ) );
       $cid->setAttribute( 'referrer', 'true' );
       $cid->setAttribute( 'offset', $referer->[1] );
       my @files = $referer->[0]->getChildren( 'content-file' );
       foreach my $file(@files) {
         $cid->addChild( XmlNode->new( 'content-file', 'content' => $file->getText(), 'attributes' => {'size' => $file->getAttribute('size') } ) ) if $file;
       }
       return $cid;
     }
     else{
       printToLog( "The path  '$referer->[2]' is not exists. Skip it." );
     }
   }
   return undef;
}

sub makeSiteAppItemNode {
    my ( $self, $parent, $ptrParams ) = @_;
    my $appItemNode = XmlNode->new('sapp-item');
    $appItemNode->setAttribute( 'enabled',
      ( $ptrParams->{'disabled'} eq 'true' ) ? 'false' : 'true' )
      if defined $ptrParams->{'disabled'};
    my $appSpecNode = XmlNode->new('sapp-spec');
    $appSpecNode->addChild( XmlNode->new( 'sapp-name', 'content' => $ptrParams->{'sapp_name'} ) );
    $appSpecNode->addChild( XmlNode->new( 'sapp-version',  'content' => $ptrParams->{'sapp_version'} ) )  if defined $ptrParams->{'sapp_version'};
    $appSpecNode->addChild( XmlNode->new( 'sapp-release', 'content' => $ptrParams->{'sapp_release'} ) )   if defined $ptrParams->{'sapp_release'};
    $appItemNode->addChild($appSpecNode);
    $appItemNode->addChild( XmlNode->new( 'license-type',
        'content' => defined( $ptrParams->{'license_type_id'} )
        ? $ptrParams->{'license_type_id'}
        : "0" ) );
    $appItemNode->addChild( XmlNode->new( 'shared', 'content' => $ptrParams->{'shared'} ) ) if defined $ptrParams->{'shared'};
    $appItemNode->addChild( XmlNode->new( 'description', 'content' => $ptrParams->{'description'} ) ) if defined $ptrParams->{'description'};
    $appItemNode->addChild( XmlNode->new( 'instances-limit', 'content' => $ptrParams->{'instances_limit'} ) ) if defined $ptrParams->{'instances_limit'};

    $parent->addChild($appItemNode);
}

sub addTemplateItem {
    my ( $node, $name, $value ) = @_;
    $node->addChild( XmlNode->new( "template-item",
        "attributes" => { "name" => $name },
        "content"    => $value ) );
}

sub makeLogrotationNode {
    my ( $self, $parent, $logRotationPtr ) = @_;

    my %logRotation = %{$logRotationPtr};

    if (%logRotation) {
      my $item = XmlNode->new( 'logrotation', 'attributes' => {
          'max-number-of-logfiles' => $logRotation{'max_number_of_logfiles'},
          'compress'               => $logRotation{'compress_enable'}
      } );

      $item->setAttribute( 'email', $logRotation{'email'} )
        unless $logRotation{'email'} eq '';
      $item->setAttribute( 'enabled', $logRotation{'turned_on'} );

      my $rottype;
      if ( $logRotation{'period_type'} eq 'by_time' ) {
        $rottype =
          XmlNode->new( 'logrotation-period',
          'attributes' => { 'period' => $logRotation{'period'} } );
      }
      else {
        $rottype =
          XmlNode->new( 'logrotation-maxsize',
          'content' => $logRotation{'period'} );
      }
      $item->addChild($rottype);
      $parent->addChild($item);
    }
}

sub makeDatabaseNode {
    my ( $self, $dbId, $dbName, $dbType, $parent, $parentId, $optionalPtr, $dbServerPtr, $dbUsersPtr, $contentDescriptionPtr ) = @_;

    my %optional = %{$optionalPtr};
    my %dbServer = %{$dbServerPtr};
    my @dbUsers = @{$dbUsersPtr};
    my %contentDescription = %{$contentDescriptionPtr};

    my $root = XmlNode->new( 'database', 'attributes' => {
        'name' => $dbName,
        'type' => $dbType
        } );

    $root->setAttribute( 'version', $optional{'version'} )
      if $optional{'version'};
    $root->setAttribute( 'charset', $optional{'charset'} )
      if $optional{'charset'};
    $root->setAttribute( 'version', $optional{'version'} )
      if $optional{'version'};

    if (%dbServer) {
      my $dbServerNode = XmlNode->new('db-server');
      $dbServerNode->setAttribute( 'type', $dbServer{'type'} );
      $dbServerNode->addChild( XmlNode->new( 'host', 'content' => $dbServer{'host'} ) );
      $dbServerNode->addChild( XmlNode->new( 'port', 'content' => $dbServer{'port'} ) );

      $root->addChild($dbServerNode);
    }

    my $item;
    foreach my $user (@dbUsers) {
      my $dbUserRoot = XmlNode->new( 'dbuser',
        'attributes' => { 'name' => $user->{'login'} } );

      $item = makePasswordNode( $user->{'password'},
        normalizePasswordType( $user->{'type'} ) );

      $dbUserRoot->addChild($item);
      $root->addChild($dbUserRoot);
    }

    if (%contentDescription) {
      my $cid;
      my %options;
      $options{"name"} =  $contentDescription{'name'};
      $options{"type"} = $contentDescription{'type'};
      $options{"user"}     = $contentDescription{'user'};
      $options{"password"} = $contentDescription{'password'};
      $options{"host"}     = $contentDescription{'host'};
      $options{"port"}     = $contentDescription{'port'};
      $options{"plesk_7"}  = $contentDescription{'plesk_7'};
      if( $parent eq 'subdomain' ) {
        $cid = $self->addSubdomainDbTar( 'sqldump', $parentId, $dbId, %options );
      }
      else{
        $cid = $self->addDbTar( 'sqldump', $dbId, %options );
      }
      $root->getChild( 'content', 1, 1 )->addChild( $cid ) if $cid;
    }

    $self->{databaseNodes}->{$dbId} = $root;
    return $root;
}

sub makeAutoresponderNode {
    my ( $self, $autoPtr, $mailName ) = @_;

    my %auto = %{$autoPtr};

    my $root = XmlNode->new('autoresponder');

    my $value = $self->{base64}->{'ENCODE'}->( $auto{'text'} );

    my $item = XmlNode->new( 'text', 'content' => $value );
    if ( exists $auto{'charset'} && $auto{'charset'} ne '' ) {
      $item->setAttribute( 'charset', $auto{'charset'} );
    }
    $root->addChild($item);

    if ( exists $auto{'content_type'} && $auto{'content_type'} ) {
      $root->setAttribute( 'content-type', $auto{'content_type'} );
    }

    my $name = $auto{'resp_name'};
    $root->setAttribute( 'name', $name );

    if ( $auto{'resp_on'} =~ /false/ ) {
      $root->setAttribute( 'status', 'off' );
    }

    unless ( $auto{'key_where'} =~ /no/ ) {
      $value = $auto{'key_where'} . ':' . $auto{'keystr'};
      $root->setAttribute( 'require', $value );
    }

    if ( $auto{'subject'} ) {
      $value = $self->{base64}->{'ENCODE'}->( $auto{'subject'} );
      chomp $value;
      $root->setAttribute( 'subject', $value );
    }

    if ( $auto{'reply_to'} and $auto{'reply_to'} ne $mailName ) {
      $root->setAttribute( 'replyto', $auto{'reply_to'} );
    }

    #
    # forward
    #
    if ( $auto{'redirect'} ) {
      $root->setAttribute( 'redirect', $auto{'redirect'} );
    }

    #
    # end forward
    #

    $root->addChild( XmlNode->new( 'autoresponder-limit',
        'content'    => $auto{'ans_freq'},
        'attributes' => { 'name' => 'ans-freq' } ) );

    $root->addChild( XmlNode->new( 'autoresponder-limit',
        'content'    => $auto{'mem_limit'},
        'attributes' => { 'name' => 'ans-count' } ) );

    my @attach = @{$auto{'attach'}};
    if ( @attach ) {
      foreach my $file (@attach) {
        $root->addChild( XmlNode->new( 'attach', 'attributes' => { 'file' => $file } ) );
      }
    }

    return $root;
}

sub parseCustomButtonOptions {
    my ( $node, $options ) = @_;

    #CUSTOM_BUTTON_PUBLIC
    $node->setAttribute( 'visible-to-sublogins', $options & 128 ? 'true' : 'false' );

    #CUSTOM_BUTTON_INTERNAL
    $node->setAttribute( 'open-in-same-frame', $options & 256 ? 'true' : 'false' );

    my %options = (
      "domain-id"           => 1,       #CUSTOM_BUTTON_DOM_ID
      "domain-name"         => 32,      #CUSTOM_BUTTON_DOM_NAME
      "ftp-login"           => 512,     #CUSTOM_BUTTON_FTP_USER
      "ftp-password"        => 1024,    #CUSTOM_BUTTON_FTP_PASS
      "client-id"           => 2,       #CUSTOM_BUTTON_CL_ID
      "client-company-name" => 4,       #CUSTOM_BUTTON_CNAME
      "client-contact-name" => 8,       #CUSTOM_BUTTON_PNAME
      "client-email"        => 16,      #CUSTOM_BUTTON_EMAIL
      );

    while ( my ( $optname, $optmask ) = each %options ) {
      if ( $options & $optmask ) {
        $node->addChild(
          XmlNode->new( "url-option", "attributes" => { "name" => $optname } )
        );
      }
    }
}

sub makeCustomButtonNode {
    my ( $self, $parentType, $parentId, $id, $optionsPtr, $customButtonsDir, $icon ) = @_;

    my %options = %{$optionsPtr};

    my $node = XmlNode->new('custom-button');

    my @attributes = ( 'url',
      'text',
      'sort-priority',
      'interface-place',
      'conhelp' );

    for my $nodefield (@attributes) {
      $node->setAttribute( $nodefield, $options{$nodefield} ) if exists $options{$nodefield};
    }

    if ( -e $customButtonsDir . "/" . $icon ) {
      if ($icon) {
        my $proc;
        if( $parentType eq 'admin' ) { $proc = \&addAdminTar; }
        elsif( $parentType eq 'clients' ) { $proc = \&addClientTar; }
        elsif( $parentType eq 'domains' ) { $proc = \&addDomainTar; }
        elsif( $parentType eq 'mails' ) { $proc = \&addMailnameTar; }
        else{
          printToError( "Error: makeCustomButtonNode: unsupported parent type '$parentType' ");
        }
        if( $proc ) {
          my $cid_icon =
            ( AgentConfig::tarVersionMoreThan15() )
            ? $self->$proc(
            'icon',
            $parentId,
            $icon . "." . $id,
            "directory" => $customButtonsDir,
            "add_file"  => $icon
            )
            : $self->$proc(
            'icon',
            $parentId,
            $icon . "." . $id,
            "directory" => $customButtonsDir,
            "include"   => [$icon]
            );

          $node->getChild( 'content', 1, 1 )->addChild( $cid_icon ) if $cid_icon;
        }
      }
    }
    parseCustomButtonOptions( $node, $options{'options'} );

    return $node;
}

#-----------------------------------------------------------------
#    Returns <dns-zone-param> xml node
#-----------------------------------------------------------------
sub makeDnsZoneParam( $ $ $ ) {
  my ( $self, $name, $db_unit, $db_value ) = @_;

  my $dnsZoneParam =
    XmlNode->new( 'dns-zone-param',
    'attributes' => { 'name' => $name, 'value' => $db_value } );

  my %units = (
    'second'  => 1,
    'minutes' => 60,
    'hours'   => 3600,
    'days'    => 86400,
    'weeks'   => 604800
  );
  my $string_unit;

  foreach my $quant ( keys %units ) {
    if ( $units{$quant} == $db_unit ) {
      $string_unit = $quant;
      last;
    }
  }
  $dnsZoneParam->setAttribute( 'unit', $string_unit );

  return $dnsZoneParam;
}

sub makeDnsZone {
  my ( $self, $parent, $paramsPtr, $recordsPtr ) = @_;

  my %params = %{$paramsPtr};
  my @records = @{$recordsPtr};

  my $email;
  if ( defined($params{'email'}) and (length( $params{'email'} ) != 0) ) {
    $email = $params{'email'};
  }
  else {

    # should assume some default email or something...
    $email = 'root@localhost.localdomain';
  }

  my $dnsZone = XmlNode->new( 'dns-zone',
    'attributes' => { 'email' => $email, 'type' => $params{'type'} },
    'children' => [ Status::make( $params{'status'} ) ] );

  if ( exists $params{'serial_format'} ){
     $dnsZone->setAttribute( 'serial-format', $params{'serial_format'} );
  }

  foreach my $zoneParam ( 'ttl', 'refresh', 'retry', 'expire', 'minimum' ) {
    $dnsZone->addChild(
      $self->makeDnsZoneParam(
        $zoneParam, $params{ $zoneParam . '_unit' },
        $params{$zoneParam}
      )
    );
  }

  # dns records
  for my $hash (@records) {
    my $dnsrec = $self->makeDnsRecord( $hash );
    if ($dnsrec) {
      $dnsZone->addChild($dnsrec);
    }
  }

  $parent->addChild($dnsZone);
}

#-----------------------------------------------------------------
#    Returns <dnsrec> xml node or false if given record should not be dumped
#-----------------------------------------------------------------
sub makeDnsRecord( $ ) {
  my ( $self, $hashPtr ) = @_;

  unless (ref($hashPtr) =~ /HASH/)
  {
    printToError('Error: makeDnsRecord: hashPtr is not hash ptr');
    return undef;
  }
  my %hash = %{$hashPtr};

  return if $hash{'type'}        =~ /none/;
  return if $hash{'displayHost'} && $hash{'displayHost'} =~ /.*_domainkey.*/;

  my $host =
    defined $hash{'displayHost'}
    ? $hash{'displayHost'}
    : $hash{'host'};
  my $val =
    defined $hash{'displayVal'}
    ? $hash{'displayVal'}
    : $hash{'val'};

  if ( $hash{'type'} eq 'SRV' ) {
    $host =~ s/\.$//;
    $val  =~ s/\.$//;
  }

  my ($item) =
    XmlNode->new( 'dnsrec',
    'attributes' => { 'type' => $hash{'type'}, 'src' => $host } );
  $item->setAttribute( 'dst', $val ) if ($val);
  $item->setAttribute( 'opt', $hash{'opt'} ) if ( defined $hash{'opt'} );

  return $item;
}

sub addUrlDecodedTextNode {
  my ( $parent, $name, $value ) = @_;

  $parent->addChild( XmlNode->new( $name, "content" => urlDecode($value) ) )
    if $value;
}

sub makeSysUser {
  my ( $self, $sysuserPtr ) = @_;

  my %sysuser = %{$sysuserPtr};

  #
  # attributes
  #

  my $root =
    XmlNode->new( 'sysuser',
    'attributes' => { 'name' => lc( $sysuser{'login'} ) } );
  $root->setAttribute( 'shell', $sysuser{'shell'} ) if $sysuser{'shell'} and $sysuser{'shell'} ne '/bin/false';

  my $quota;
  if ( $sysuser{'quota'} ) {
    $quota = $sysuser{'quota'};
    $root->setAttribute( 'quota', $quota ) if $quota;
  }

  #
  # end attributes
  #

  #
  # password node
  #
  $root->addChild( makePasswordNode( $sysuser{'passwd'}, $sysuser{'passwdType'} ) );

  #
  # end password node
  #

  #
  # crontab node
  #
  if ( $sysuser{'cron'} ) {
    $root->addChild( XmlNode->new( 'cron', 'content' => $sysuser{'cron'} ) );
  }

  #
  # end crontab node
  #
  return $root;
}

sub makeSappNode {
  my ( $self, $sappId, $name, $version, $release, $description, $isCommercial, $isIntegrated,
            $prefix, $isSsl, $licenseType ) = @_;

  my $sapp_installed = XmlNode->new('sapp-installed');

  #-----------------------------------------------------------------
  # sapp-spec
  #-----------------------------------------------------------------
  my $sapp_spec = XmlNode->new('sapp-spec');
  $sapp_spec->addChild( XmlNode->new( 'sapp-name', 'content' => $name ) );
  if ( $version ) {
    $sapp_spec->addChild(
      XmlNode->new( 'sapp-version', 'content' => $version ) );
  }
  if ( $release ) {
    $sapp_spec->addChild(
      XmlNode->new( 'sapp-release', 'content' => $release ) );
  }
  if ( $description ) {
    my $desc = $self->{base64}->{'ENCODE'}->( $description );
    $sapp_spec->addChild(
      XmlNode->new(
        'sapp-description',
        'attributes' => { 'encoding' => 'base64' },
        'content'    => $desc
      )
    );
  }
  if ( $isCommercial ) {
    $sapp_spec->addChild( XmlNode->new('sapp-commercial') );
  }
  if ( $isIntegrated ) {
    $sapp_spec->addChild( XmlNode->new('sapp-integrated') );
  }
  $sapp_installed->addChild($sapp_spec);

  if ( defined($licenseType) ) {
    $sapp_installed->addChild(
      XmlNode->new(
        'sapp-installed-license-type',
        'content' => $licenseType
      )
    );
  }

  return $sapp_installed;
}

sub makeSiteAppInstalled{
  my ( $self, $sappId, $prefix, $isSsl ) = @_;

  my $sapp_installed = $self->{sappNodes}->{$sappId};

  # sapp-installdir
  my $sapp_installdir =
    XmlNode->new( 'sapp-installdir',
    'children' => [ XmlNode->new( 'sapp-prefix', 'content' => $prefix ) ] );
  if ( $isSsl ) {
    $sapp_installdir->addChild( XmlNode->new('sapp-ssl') );
  }
  $sapp_installed->addChild($sapp_installdir);
}

sub makeAnonftpPermissionNode {
  my ( $parent, $name ) = @_;

  $parent->addChild(
    XmlNode->new( 'anonftp-permission', "attributes" => { "name" => $name } ) );
}

sub makeAnonftpLimitNode {
  my ( $parent, $name, $value ) = @_;
  if ( $value != 0 ) {
    $parent->addChild(
      XmlNode->new(
        "anonftp-limit",
        "attributes" => { "name" => $name },
        "content"    => $value
      )
    );
  }
}

sub blockToNum {
  my ( $self, $mask ) = @_;
  my $longMask = unpack( "N", pack( "C4", split( /\./, $mask ) ) );
  my $block;
  for ( $block = 0 ; $block < 32 ; $block = $block + 1 ) {
    my $tmp = 2**( 32 - $block - 1 );
    last if !( $longMask & $tmp );
  }
  return $block;
}

sub makeSystemIpNode {
  my ( $self, $parent, $ptrIp ) = @_;
  my $correctMask = $self->blockToNum( $ptrIp->{'mask'} );
  $parent->getChild( 'properties', 1 )->addChild(
    XmlNode->new(
      'system-ip',
      'attributes' => { 'certificate' => $ptrIp->{'certificate'} },
      'children'   => [
        XmlNode->new(
          'ip',
          'children' => [
            XmlNode->new( 'ip-type', 'content' => "$ptrIp->{'type'}" ),
            XmlNode->new(
              'ip-address', 'content' => $ptrIp->{'ip_address'}
            )
          ]
        ),
        XmlNode->new( 'ip-netmask',   'content' => "$correctMask" ),
        XmlNode->new( 'ip-interface', 'content' => "$ptrIp->{'iface'}" )
      ]
    )
  );
}

sub makeTemplateNode {
  my ( $self, $nodeName, $templateName, $templateNote, $templatePtr, $logRotationPtr, $ipPoolPtr ) = @_;

  my @template = @{$templatePtr};
  my %logRotation = %{$logRotationPtr};
  my %ipPool = %{$ipPoolPtr};

  my $node =
    XmlNode->new( $nodeName,
    'attributes' => { 'name' => $templateName } );

  if ( $templateNote ) {
    $node->setAttribute( 'note-id', $templateNote );
  }

  my %template;

  for my $ptrRow ( @{$templatePtr} ) {
      #print "************ $ptrRow with $templatePtr " . ref($ptrRow) . "\n";;
      my ( $element, $value ) = @{$ptrRow};
      addTemplateItem( $node, $element, $value );
 }

  if (%logRotation){
      $self->makeLogrotationNode( $node, \%logRotation );
  }

  if (%ipPool and scalar(keys %ipPool) > 0){
	  my $ipPoolNode = XmlNode->new( 'ip_pool' );

	  for my $ip ( keys %ipPool ) {
        $ipPoolNode->addChild(
	      $self->makeIpNode($ip, $ipPool{$ip} )
	    );
	  }

      $node->addChild($ipPoolNode);
  }

  return $node;
}

sub makeIpNode {
  my ($self, $ip, $iptype) = @_;
  return XmlNode->new(
    'ip',
    'children' => [
      XmlNode->new( 'ip-type',    content => $iptype ),
      XmlNode->new( 'ip-address', content => $ip )
    ]
  );
}

sub setServerVirusfilter {
  my ($self, $paramsPtr) = @_;

  my $root = $self->{serverNode};

  my $content = defined($paramsPtr->{'virusfilter'}) ? $paramsPtr->{'virusfilter'} : "none";

  $root->addChild(XmlNode->new('virusfilter','attributes' => { 'state' => 'inout' }, 'content' => $content) );
}

1;
