package PleskStructure;

use PleskVersion;
use XmlNode;

# IP address is the pair (address, type)

my %ipType;                #ip -> 'shared'|'exclusive'
my %clientDomains;         #clientName -> domainName
my %clientIds;		   #clientId -> clientName
my %clientChilds;	   #clientId -> clientName
my %domainIp;              #domainName -> ip
my %domainIds;              #domainName -> id
my $defaultIp;
my %clientData; #clientName -> { 'id'-> $id, 'type' -> $type, 'parent' -> $parent, 'ip' -> @ip, 'guid' -> $guid }
my $adminName;

my $dbh;

## public interface

sub fastInit {
  $dbh = shift;
  if (PleskVersion::atLeast(9, 0, 0)) {
    __populateIpsFromIPAddressesTable9();
  }
  elsif( PleskVersion::atLeast(6, 0, 0) ){
    __populateIpsFromIPAddressesTable();
  }
  else{
    __getDefaultIp();
  }
}

sub init {
  $dbh = shift;

  if (PleskVersion::atLeast(9, 0, 0)) {
    __init90();
  } elsif (PleskVersion::atLeast(6, 0, 0)) {
    __init60();
  } elsif (PleskVersion::atLeast(5, 0, 0)) {
    __init50();
  } else {
    __init25();
  }
}

sub getClients {
  my( @ret );
  foreach my $clientName( keys %clientData ){
    push @ret, $clientName if ${clientData}{$clientName}{'type'} eq 'client';
  }
  return @ret;
}

sub getResellers {
  my( @ret );
  foreach my $clientName( keys %clientData ){
    push @ret, $clientName if ${clientData}{$clientName}{'type'} eq 'reseller';
  }
  return @ret;
}

sub getResellersFromId{
  my ( $ids ) = @_;
  my ( @ret, @wrong );
  foreach my $id( @{$ids} ){
     my $login = getClientNameFromId( $id );
     if( $login ){
        if( getClientType( $login ) eq 'reseller' ) {  push @ret, $login;  }
        else{  push @wrong, $id; }
     }
     else{ push @wrong, $id; }
  }
  @{$ids} = @wrong;
  return @ret;
}

sub getClientsFromId{
  my( $ids )= @_;
  my ( @ret, @wrong );
  foreach my $id( @{$ids} ){
     my $login = getClientNameFromId( $id );
     if( $login ){
        if( getClientType( $login ) eq 'client' ) {  push @ret, $login;  }
        else{  push @wrong, $id; }
     }
     else{ push @wrong, $id; }
  }
  @{$ids} = @wrong;
  return @ret;
}

sub getDomainsFromId{
  my ( $ids ) = @_;
  my ( @ret, @wrong );
  foreach my $id( @{$ids} ){
     my $domain = getDomainNameFromId( $id );
     if( $domain ){
        push @ret, $domain;
     }
     else{ push @wrong, $id; }
  }
  @{$ids} = @wrong;
  return @ret;
}

sub getAdminClients {
  if (PleskVersion::atLeast(9, 0, 0)) {
    my @ret = getMyClients( $adminName );
    #push @ret, $adminName;
    return @ret;
  }
  else{
    return getClients();
  }
}

sub getAdminDomains {
  if (PleskVersion::atLeast(9, 0, 0)) {
     return getDomainsForClient( $adminName );
  }
  else{
    my @ret;
    return @ret;
  }
}

sub getAdminId{
  return getClientId( $adminName ) if $adminName;
  return undef;
}

sub getAdminGuid{
  return getClientGuid( $adminName ) if $adminName;
  return "00000000-0000-0000-0000-000000000000";
}

sub getClientsCount{
  my ( $parentName ) = @_;
  my $count = 0;
  if( $parentName ){
     my $parentId = $clientData{$parentName}{'id'};
     foreach my $clientName( keys %clientData ){
       if( $clientData{$clientName}{'parent'} == $parentId ){
          $count += 1;
          if( $clientData{$clientName}{'type'} eq 'reseller' ){
            $count += getClientsCount( $clientData{'id'} );
          }
       }
     }
  }
  else{
    $count = scalar( keys %clientData );
  }
  return $count;
}

sub getDomainsCount{
  my ($parentName ) = @_;
  my $count = 0;
  if( $parentName ){
     my $parentId = $clientData{$parentName}{'id'};
     foreach my $clientName( keys %clientData ){
       if( $clientData{$clientName}{'parent'} == $parentId ){
          $count += scalar( getDomainsForClient( $clientName ) );
          if( $clientData{$clientName}{'type'} eq 'reseller' ){
            $count += getDomainsCount( $clientData{'id'} );
          }
       }
       elsif( $clientData{$clientName}{'id'} == $parentId ){
         $count += scalar( getDomainsForClient( $clientName ) );
       }
     }
  }
  else{
    foreach my $clientName( keys %clientDomains ){
      $count += scalar( getDomainsForClient( $clientName ) )
    }
  }
  return $count;
}

sub PleskStructure::sortClients{
  my ( $a, $b ) = @_;
  return $a cmp $b if $clientData{$a}{'type'} eq $clientData{$b}{'type'};
  return  1 if $clientData{$a}{'type'} eq 'reseller';
  return -1 if $clientData{$b}{'type'} eq 'reseller';
  return  1 if $clientData{$a}{'type'} eq 'admin';
  return -1 if $clientData{$b}{'type'} eq 'admin';
  die "Cannot compare clients '$a', '$b'. Invalid types!";
}


sub getMyClients {
  my ( $parentName ) = @_;
  my( @ret );
  die "Cannot find client '$parentName'!"  if not exists $clientData{$parentName};
  my $parentId = $clientData{$parentName}{'id'};
  return @{$clientChilds{$parentId}} if exists $clientChilds{$parentId};
  return @ret;
}

sub getClientType{
  my( $clientName ) = @_;
  die "Cannot determine client type. Cannot find client '$clientName'!"  if not exists $clientData{$clientName};
  my $ret = $clientData{$clientName}{'type'};
  $ret = 'client' if not $ret;
  return $ret;
}

sub getDomainNameFromId{
  my( $domainId ) = @_;
  my ( $name, $data );
  my $ret = '';
  while ( ($name, $data) = each %domainIds) {
    $ret = $name if $data == $domainId;
  }
  return $ret;
}


sub getClientNameFromId{
  my( $clientId ) = @_;
  return $clientIds{$clientId};
}

sub getClientParentId{
  my( $clientName ) = @_;
  return  $clientData{$clientName}{'parent'};
}

sub getClientId{
  my( $clientName ) = @_;
  return  $clientData{$clientName}{'id'};
}

sub getDomainId{
  my( $domainName ) = @_;
  return  $domainIds{$domainName};
}

sub getClientGuid{
  my( $clientName ) = @_;
  return  $clientData{$clientName}{'guid'};
}

sub getDomains {
  return keys %domainIp;
}

sub getDomainsForClient {
  my ($clientName) = @_;
  return @{ $clientDomains{$clientName} };
}

sub getClientIps {
  my ($clientName) = @_;
  return %{$clientData{$clientName}{'ip'}};
}

sub getClientIpType {
  my ($clientName, $ip) = @_;
  return ${$clientData{$clientName}{'ip'}}{$ip} if exists ${$clientData{$clientName}{'ip'}}{$ip};
  return $ipType( $ip );
}

sub getDomainIp {
  my ($domainName) = @_;
  return $domainIp{$domainName};
}

sub isExclusiveIp {
  my ($ip) = @_;

  return $ipType{$ip} eq "exclusive";
}

sub getDefaultIp{
  return $defaultIp if $defaultIp;
  return '127.0.0.1';
}

## Plesk 2.5

sub __init25 {
  my $defaultIp = __getDefaultIp();
  __populateIpsFromDns();

  my $sql = 'SELECT d.name, r.val FROM dns_recs r, domains d WHERE r.dom_id = d.id AND r.type = "A"';
  $dbh->execute($sql);
  my $row;
  while ( $row = $dbh->fetchrow() ) {
    my ( $domainName, $ipAddress ) = @{$row};
    $ipAddress ||= $defaultIp;
    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  __populateClientsFromDomains();
}

## Plesk 5.0, 5.5

sub __init50 {
  my $defaultIp = __getDefaultIp();
  __populateIpsFromDns();
  __populateIpsFromList("SELECT DISTINCT(ip_address) FROM hosting");

  my $sql = "SELECT d.name, h.ip_address FROM domains d LEFT JOIN hosting h ON h.dom_id = d.id";
  $dbh->execute($sql);
  my $row;
  while ( $row = $dbh->fetchrow() ) {
    my ( $domainName, $ipAddress ) = @{$row};
    $ipAddress ||= $defaultIp;

    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  __populateClientsFromDomains();
}


## Plesk 9.0 -

sub __init90 {
  __populateIpsFromIPAddressesTable9();
  __populateClientIpPool9();
  __populateDomains();
}

sub __populateIpsFromIPAddressesTable9 {
  my $row;
  my $def_ip_id;
  my $sql = "SELECT val from misc where param='def_ip_id'";
  $dbh->execute($sql);
  if( $row = $dbh->fetchrow() ) {
    $def_ip_id = $row->[0];
  }
  $dbh->finish();

  my $sql = "SELECT ip_address, id FROM IP_Addresses";
  $dbh->execute($sql);
  while ( $row = $dbh->fetchrow() ) {
    $ipType{ $row->[0] } = 'shared';
    $defaultIp = $row->[0] if $row->[1] eq $def_ip_id;
  }
  $dbh->finish();
}


sub __populateClientIpPool9 {
  my $sql = "SELECT c.id, c.login, p.ip_address, c.type, c.parent_id, pl.type, c.guid FROM clients c "
    . "LEFT JOIN ip_pool pl ON c.pool_id = pl.id "
    . "LEFT JOIN IP_Addresses p ON pl.ip_address_id = p.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($id, $clientName, $ipAddress, $cltype, $clparent, $iptype, $guid ) = @{$row};
    $clientIds{$id} =  $clientName;
    $clientData{$clientName} = { 'id' => $id, 'type' => $cltype, 'parent' => $clparent, 'ip' => {}, 'guid' => $guid } if not exists $clientData{$clientName};
    ${$clientData{$clientName}{'ip'}}{$ipAddress} = $iptype if ($ipAddress);
    if( $cltype eq 'admin' ){
       $adminName = $clientName;
       $ipType{$ipAddress} = $iptype;
    }
  }
  foreach my $clientName( keys %clientData ){
     my $clparent = $clientData{$clientName}{'parent'};
     if( $clparent ) {
        $clientChilds{$clparent} = () if not exists $clientChilds{$clparent};
        push @{$clientChilds{$clparent}}, $clientName;
     }
  }
  $dbh->finish();
}

## Plesk 6.0 - 8.0

sub __init60 {
  __populateIpsFromIPAddressesTable();
  __populateClientIpPool();
  __populateDomains();
}

## common

sub __getDefaultIp {
  unless ($defaultIp) {
    my $sql = "SELECT val FROM misc WHERE param = 'NameVirtualHost'";
    $dbh->execute($sql);
    my @row = @{ $dbh->fetchrow() };
    $defaultIp = $row[0];
    $dbh->finish();
  }
  return $defaultIp;
}

sub __populateIpsFromList {
  my ($sql) = @_;

  $ipType{$defaultIp} = "shared";

  $dbh->execute($sql);
  my $row;
  while ( $row = $dbh->fetchrow() ) {
    my @row = @{$row};
    my ($ip) = $row[0];
    $ipType{$ip} = $ip eq $defaultIp ? "shared" : "exclusive";
  }
  $dbh->finish();
}

sub __populateIpsFromDns {
  __populateIpsFromList("SELECT DISTINCT(val) FROM dns_recs WHERE type = 'A'");
}

sub __populateClientsFromDomains {
  my %clientToIp;

  my $sql = "SELECT c.id, c.login, d.name, d.id dom_id FROM clients c, domains d WHERE d.cl_id = c.id";
  $dbh->execute($sql);
  my $row;
  while ( $row = $dbh->fetchrow() ) {
    my ( $id, $clientName, $domainName, $domainId ) = @{$row};

    if ( !exists( $clientDomains{$clientName} ) ) {
      $clientDomains{$clientName} = [];
    }
    push @{ $clientDomains{$clientName} }, $domainName;

    $clientData{$clientName} = { 'id' => $id, 'type' => 'client', 'parent' => 0, 'ip' => {} } if not exists $clientData{$clientName};
    $clientIds{$id} =  $clientName;

    if ( !exists( $clientToIp{$clientName} ) ) {
      $clientToIp{$clientName} = {};
    }
    $clientToIp{$clientName}->{ $domainIp{$domainName} } = $ipType{$domainIp{$domainName}};
    $domainIds{$domainName} = $domainId;
  }
  $dbh->finish();

  for my $clientName (keys %clientToIp) {
     %{$clientData{$clientName}{'ip'}} = %{$clientToIp{$clientName}};
  }
}

sub __populateIpsFromIPAddressesTable {
  my $row;
  my $def_ip_id;
  my $sql = "SELECT val from misc where param='def_ip_id'";
  $dbh->execute($sql);
  if( $row = $dbh->fetchrow() ) {
    $def_ip_id = $row->[0];
  }
  $dbh->finish();

  my $sql = "SELECT ip_address, type, id FROM IP_Addresses";
  $dbh->execute($sql);
  while ( $row = $dbh->fetchrow() ) {
    $ipType{ $row->[0] } = $row->[1];
    $defaultIp = $row->[0] if $row->[2] eq $def_ip_id;
  }
  $dbh->finish();
}

sub __populateClientIpPool {
  my $sql = "SELECT c.id, c.login, p.ip_address FROM clients c "
	    . "LEFT JOIN Repository r ON c.pool_id = r.rep_id "
	    . "LEFT JOIN IP_Addresses p ON r.component_id = p.id";
  $dbh->execute($sql);
  my $row;
  while ($row = $dbh->fetchrow()) {
    my ($id, $clientName, $ipAddress) = @{$row};
    $clientData{$clientName} = { 'id' => $id, 'type' => 'client', 'parent' => 0, ip => {} } if not exists $clientData{$clientName};
    ${$clientData{$clientName}{'ip'}}{$ipAddress} = $ipType{$ipAddress} if $ipAddress;
    $clientIds{$id} =  $clientName;
  }
    $dbh->finish();
}

sub __populateDomains {
  my $domainNameColumn = 'name';
  if ( PleskVersion::atLeast( 7, 1, 0 ) ) {
    $domainNameColumn = 'displayName';
  }

  my $sql = "SELECT d.$domainNameColumn, p.ip_address FROM domains d, IP_Addresses p, hosting h "
           . "WHERE p.id = h.ip_address_id AND h.dom_id = d.id";
  $dbh->execute($sql);
  my $row;
  while ( $row = $dbh->fetchrow() ) {
    my ( $domainName, $ipAddress, $domainId ) = @{$row};
    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  $sql = "SELECT d.$domainNameColumn, p.ip_address FROM domains d, IP_Addresses p, forwarding f "
        . "WHERE p.id = f.ip_address_id AND f.dom_id = d.id";
  $dbh->execute($sql);
  while ( $row = $dbh->fetchrow() ) {
    my ( $domainName, $ipAddress ) = @{$row};
    $domainIp{$domainName} = $ipAddress;
  }
  $dbh->finish();

  $sql = "SELECT d.$domainNameColumn FROM domains d WHERE htype = 'none'";
  $dbh->execute($sql);
  while ( $row = $dbh->fetchrow() ) {
    $domainIp{ $row->[0] } = undef;
  }
  $dbh->finish();

  $sql = "SELECT d.$domainNameColumn, c.login, d.id dom_id FROM clients c LEFT JOIN domains d ON d.cl_id = c.id";
  $dbh->execute($sql);
  while ( $row = $dbh->fetchrow() ) {
    my ( $domainName, $clientName, $domainId ) = @{$row};
    $domainIds{$domainName} = $domainId;
    if ( !exists( $clientDomains{$clientName} ) ) {
      $clientDomains{$clientName} = [];
    }
    if ($domainName) {
      push @{ $clientDomains{$clientName} }, $domainName;
    }
  }
  $dbh->finish();
}




1;

# Local Variables:
# mode: cperl
# cperl-indent-level: 2
# indent-tabs-mode: nil
# tab-width: 4
# End:
