package PleskVersion;

use strict;

require PleskMaxVersion;

my ($major, $minor, $minorMinor);

sub init {
  my ($productRootD) = @_;

  my $versionFile = (-d $productRootD) ? "$productRootD/version" : "/usr/local/psa/version";

  open(VERSION, "< $versionFile") || die "Unable to find Plesk versionfile: $versionFile";
  binmode (VERSION);
  my $out = <VERSION>;
  chomp $out;
  close(VERSION);

  if ($out =~ /^(\d+)\.(\d+)(.*)$/) {
    $major = $1;
    $minor = $2;

    if ($3 =~ /^\.(\d+)/) {
      $minorMinor = $1;
    } else {
      $minorMinor = 0;
    }
  }
}

sub get {
  if ($major) {
    return "$major.$minor.$minorMinor";
  }
}

# 2.5 <= ver <= last known plesk version
sub isSupported {
  return 0 if $major < 2 || ($major == 2 && $minor < 5);

  return 0 if $major > $PleskMaxVersion::major;
  return 0 if $major == $PleskMaxVersion::major and $minor > $PleskMaxVersion::minor;
  return 0 if $major == $PleskMaxVersion::major and $minor == $PleskMaxVersion::minor
    and $minorMinor > $PleskMaxVersion::minorMinor;

  return 1;
}

sub atLeast {
  my ($lim_major, $lim_minor, $lim_minorMinor) = @_;

  return 1 if $major > $lim_major;
  return 1 if $major == $lim_major and $minor > $lim_minor;
  return 1 if $major == $lim_major and $minor == $lim_minor and $minorMinor >= $lim_minorMinor;

  return 0;
}

1;
