import backup

import os, sys, getopt, xml_sax

def getBriefInfo(outFile, dirName, files, schema, params):
    data = {}

    for f in files:
        fileName = os.path.join(dirName, f)

        try:
            backupFile = backup.backupFileFactory(schema, fileName, **params)
            if not backupFile.exists():
                continue
        except Exception:
            continue

        fileInfo = {}

        try:
            size = backupFile.size()
            if size:
                fileInfo['size'] = str(size)
        except IOError, e:
            pass
        except OSError, e:
            pass

        try:
            time = backupFile.time()
            if time:
                fileInfo['time'] = str(time)
        except IOError:
            pass
        except OSError:
            pass

        b = backup.Backup(schema, fileName, **params)

        try:
            description = b.getMetadata().getDescription()
            if description:
                fileInfo['description'] = description
        except Exception:
            pass

        # explicit close to save FTP connections (bug #110177)
        b.close()

        data[f] = fileInfo

    writer = xml_sax.myXMLGenerator(outFile, "UTF-8")
    writer.startDocument()
    writer.startElement("directory-info", {})
    for f, d in data.items():
        writer.startElement(d.has_key("description") and "backup" or "file", d)
        # convert to unicode from latin1, as FTP server replies in unicode and XML should be UTF8
        writer.characters(f.decode('latin1'))
        writer.endElement(d.has_key("description") and "backup" or "file")
    writer.endElement("directory-info")
    writer.endDocument()


def getDetailedInfo(outFile, fileName, schema, params):
    b = backup.Backup(schema, fileName, **params)
    info = b.getMetadata().getInfo()

    writer = xml_sax.myXMLGenerator(outFile, "UTF-8")
    info.serialize(writer)


def usage(exit_code = 1):
    print "Usage:"
    print "backup-info brief [--ftp-host=<host> --ftp-login=<login>] <dir>"
    print "backup-info detailed [--ftp-host=<host> --ftp-login=<login>] <dir>/<file>"
    print "backup-info help"
    print ""
    print "Note: Password for FTP access should be specified in"
    print "      environment variable 'FTP_PASSWORD'"
    print "Note: Brief mode waits filelist from stdin"
    sys.exit(exit_code)

def main():
    if len(sys.argv) < 2:
        usage()

    command = sys.argv[1]
    if command == "help":
        usage(0)

    if command != "brief" and command != "detailed":
        usage()

    try:
        opts, args = getopt.getopt(sys.argv[2:], "", ["ftp-host=", "ftp-login="])
    except getopt.GetoptError, e:
        usage()

    if len(args) < 1:
        usage()

    ftpHost = None
    ftpLogin = None
    ftpPassword = None

    for o, a in opts:
        if o == "--ftp-host":
            if ftpHost != None:
                usage()
            ftpHost = a
        if o == "--ftp-login":
            if ftpLogin != None:
                usage()
            ftpLogin = a

    params = {}
    if ftpHost:
        schema = 'ftp'
        if not(ftpLogin):
            usage()
        try:
            ftpPassword = os.environ["FTP_PASSWORD"]
        except KeyError:
            usage()
        params['host'] = ftpHost
        params['login'] = ftpLogin
        params['password'] = ftpPassword
    else:
        if ftpLogin:
            usage()
        schema = 'file'

    if command == "brief":
        if len(args) > 1:
            files = args[1:]
        else:
            files = sys.stdin.read().split("\0")
        getBriefInfo(sys.stdout, args[0], files, schema, params)
    else:
        getDetailedInfo(sys.stdout, args[0], schema, params)


if __name__ == '__main__':
    main()
