from __future__ import generators
import commands
import getopt
import sys
import filters
import dump_walker
import backup_walker
import backup


class ListDumpVisitor(object):
    def indentOutput(self, level):
        sys.stdout.write("  " * (2 *level))

    def visit(self, info, x, parents):
        self.indentOutput(len(parents))
        print "[" + info.__repr__() + "]"


class ListBackupVisitor(object):
    def __init__(self, filter, verbosity):
        self.filter = filter
        self.verbosity = verbosity

    def visitMetadata(self, metadata):
        def processMetadata(fh):
            dump_walker.walkDump(filters.FilterVisitor(ListDumpVisitor(), self.filter, self.verbosity), fh)
        metadata.process(processMetadata)

        raise backup_walker.StopProcessing()


class ListCommand(commands.HelpProvider):
    USAGE_STRING = "list [-n|--no-recursive] [objects] <backup-file>"
    SHORT_HELP_STRING = "Lists the content of backup"
    HELP_STRING = """Lists the index of backup, or selected parts of backup.

Selecting parts of backup is described in 'backup-unpack help
select-objects'.  By default 'list' lists index recursively. This may
be disabled supplying -n|--no-recursive option.  """

    def __init__(self, verbosity):
        self.verbosity = verbosity

    def parse(self, args):
        if len(args) == 0:
            raise RuntimeError("No backup file specified")
        self.backupFile = args.pop()

        opts, args = getopt.getopt(args, "n", ["no-recursive"])
        recursive = len(opts) == 0

        self.filter = filters.parseStringFilter(args)
        if recursive:
            self.filter.setRecursive()

    def act(self):
        backupFile = backup.backupFileFactory('file', self.backupFile)
        if not backupFile.exists():
            raise RuntimeError("Backup file '" + self.backupFile + "' does not exist")

        b = backup.Backup('file', self.backupFile)

        backup_walker.walkBackup(b, ListBackupVisitor(self.filter, self.verbosity))

commands.register("list", ListCommand, 1)

# Local Variables:
# mode: python
# indent-tabs-mode: nil
# tab-width: 4
# End:
