from __future__ import generators
import xml.dom.minidom

def childElements(x):
    for n in x.childNodes:
        if n.nodeType == xml.dom.minidom.Node.ELEMENT_NODE:
            yield n

# Python 2.2 has no x.attributes.has_key() and x.attributes[key]
def getAttr(x, name):
    for i in range(x.attributes.length):
        if x.attributes.item(i).name == name:
            return x.attributes.item(i).value
    return None

def childElementsName(x, name):
    for n in childElements(x):
        if n.nodeName == name:
            yield n

def childTextNodes(x):
    for n in x.childNodes:
        if n.nodeType == xml.dom.minidom.Node.TEXT_NODE:
            yield n.data

def textContent(x):
    return "".join(childTextNodes(x))

def _ext(cid):
    if cid.endswith(":gzipped"):
        return ".gz"
    else:
        return ""

def _parseCustomButton(object, c, prefix = None):
    cid = getAttr(c, "cid-icon")
    if None != cid:
        buttonName = c.attributes["text"].value.replace("/", "_")
        if prefix:
            prefix = prefix + "-"
        else:
            prefix = ""
        object._addCid(cid, "%sicon-%s.tar%s" % (prefix, buttonName, _ext(cid)))

def _addCidIfAttribute(object, x, attributeName, fileName):
    cid = getAttr(x, attributeName)
    if None != cid:
        object._addCid(cid, fileName + _ext(cid))

class PleskObject(object):
    def __init__(self, type_, name):
        self.type_ = type_
        self.name = name
        self.cids = {}
    def __repr__(self):
        return self.type_ + " " + self.name
    def objectDir(self):
        return self.type_ + "-" + self.name

    def _addCid(self, cid, name):
        self.cids[cid] = name
    def getCids(self):
        return self.cids

class Client(PleskObject):
    def __init__(self, x):
        PleskObject.__init__(self, "client", x.attributes["name"].value)
        cid = getAttr(x, "cid-skeleton")
        if None != cid:
            self._addCid(cid, "skeleton.tar" + _ext(cid))
        for c in childElementsName(x, "custom-button"):
            _parseCustomButton(self, c)

    def objectDir(self):
        return self.name

class Domain(PleskObject):
    def __init__(self, x):
        PleskObject.__init__(self, "domain", x.attributes["name"].value)

        for n in childElements(x):
            if n.nodeName == "tomcat":
                self._parseTomcat(n)
            elif n.nodeName == "phosting":
                self._parsePhosting(n)
            elif n.nodeName == "custom-button":
                _parseCustomButton(self, n)

    def _parseTomcat(self, t):
        for w in childElementsName(t, "webapp"):
            cid = w.attributes["cid"].value
            fileName = "javaapp-%s.tar%s" % (w.attributes["name"].value, _ext(cid))
            self._addCid(cid, fileName)

    def _parsePhosting(self, p):
        _addCidIfAttribute(self, p, "cid_docroot", "httpdocs.tar")
        _addCidIfAttribute(self, p, "cid_docroot_ssl", "httpsdocs.tar")
        _addCidIfAttribute(self, p, "cid_cgi", "cgi-bin.tar")
        _addCidIfAttribute(self, p, "cid_webstat", "webstat.tar")
        _addCidIfAttribute(self, p, "cid_webstat_ssl", "webstat-ssl.tar")
        _addCidIfAttribute(self, p, "cid_ftpstat", "ftpstat.tar")
        _addCidIfAttribute(self, p, "cid_private", "private-data.tar")
        _addCidIfAttribute(self, p, "cid_conf", "conf.tar")
        _addCidIfAttribute(self, p, "cid_logs", "logs.tar")
        _addCidIfAttribute(self, p, "cid_anon_ftpstat", "anon-ftpstat.tar")
        _addCidIfAttribute(self, p, "cid_error_docs", "errordocs.tar")

        for n in childElements(p):
            if n.nodeName == "subdomain":
                self._parseSubdomain(n)
            elif n.nodeName == "anonftp":
                self._parseAnonymousFtp(n)

    def _parseSubdomain(self, s):
        name = s.attributes["name"].value
        _addCidIfAttribute(self, s, "cid_docroot", name + "-httpdocs.tar")
        _addCidIfAttribute(self, s, "cid_docroot_ssl", name + "-httpsdocs.tar")
        _addCidIfAttribute(self, s, "cid_cgi", name + "-cgi-bin.tar")
        _addCidIfAttribute(self, s, "cid_conf", name + "-conf.tar")

    def _parseAnonymousFtp(self, a):
        _addCidIfAttribute(self, a, "cid", "anonftp.tar")
        _addCidIfAttribute(self, a, "cid_incoming", "anonftp-incoming.tar")
            
    def objectDir(self):
        return self.name

class PleskObjectWithParent(PleskObject):
    def __init__(self, type_, name, parent):
        PleskObject.__init__(self, type_, name)
        self.parent = parent
    def __repr__(self):
        return PleskObject.__repr__(self) + "@" + self.parent.name
    def objectDir(self):
        return PleskObject.objectDir(self) + "@" + self.parent.name

class Mailname(PleskObjectWithParent):
    def __init__(self, x, parent):
        PleskObjectWithParent.__init__(self, "mailname",
                                       x.attributes["name"].value, parent)
        for n in childElements(x):
            if n.nodeName == "mailbox":
                self._parseMailbox(n)
            elif n.nodeName == "autoresponders":
                self._parseAutoresponders(n)
            elif n.nodeName == "spamassassin":
                self._parseSpamassassin(n)
            elif n.nodeName == "custom-button":
                _parseCustomButton(self, n)

    def _parseMailbox(self, n):
        _addCidIfAttribute(self, n, "cid", "mail.tar")
        _addCidIfAttribute(self, n, "cid_inbox", "inbox.tar")

    def _parseAutoresponders(self, n):
        _addCidIfAttribute(self, n, "cid_attaches", "attaches.tar")

    def _parseSpamassassin(self, n):
        _addCidIfAttribute(self, n, "cid", "spamassassin-db.tar")
                
    def objectDir(self):
        return self.name + "@" + self.parent.name


def _findDatabaseServer(x):
    host = None
    port = None
    for c in childElementsName(x, "db-server"):
        for cc in childElements(c):
            if cc.nodeName == "host":
                host = textContent(cc)
            if cc.nodeName == "port":
                port = textContent(cc)

    return (host, port)


class Database(PleskObject):
    def __init__(self, x):
        PleskObject.__init__(self, "database", x.attributes["name"].value)
        (self.host, self.port) = _findDatabaseServer(x)
        self.dbType = x.attributes["type"].value
        
        _addCidIfAttribute(self, x, "cid", "dump.sql")

    def __repr__(self):
        r = "database " + self.name
        if self.host:
            r += "@" + self.host
            if self.port:
                r += ":" + self.port
        if self.dbType:
            r += "/" + self.dbType
        return r
    def objectDir(self):
        r = "database-" + self.name
        if self.host:
            r += "@" + self.host
            if self.port:
                r += "-" + self.port
        if self.dbType:
            r += "-" + self.dbType
        return r

class Maillist(PleskObjectWithParent):
    def __init__(self, x, parent):
        PleskObjectWithParent.__init__(self, "maillist",
                                       x.attributes["name"].value, parent)
        _addCidIfAttribute(self, x, "cid", "archive.tar")


class Webuser(PleskObjectWithParent):
    def __init__(self, x, parent):
        PleskObjectWithParent.__init__(self, "webuser",
                                       x.attributes["name"].value, parent)
        _addCidIfAttribute(self, x, "cid", "httpdocs.tar")
        _addCidIfAttribute(self, x, "cid_private", "private-data.tar")

def walkWebuser(visitor, x, parents):
    webuser = Webuser(x, parents[-1])
    visitor.visit(webuser, x, parents)

def walkPhosting(visitor, x, parents):
    for c in childElements(x):
        if c.nodeName == "webuser":
            walkWebuser(visitor, c, parents)

def walkMaillist(visitor, x, parents):
    maillist = Maillist(x, parents[-1])
    visitor.visit(maillist, x, parents)

def walkMaillists(visitor, x, parents):
    for c in childElementsName(x, "maillist"):
            walkMaillist(visitor, c, parents)

def walkDatabase(visitor, x, parents):
    database = Database(x)
    visitor.visit(database, x, parents)

def walkMailname(visitor, x, parents):
    mailname = Mailname(x, parents[-1])
    visitor.visit(mailname, x, parents)

def walkMailsystem(visitor, x, parents):
    # historical node name. should be changed to mailname
    for c in childElementsName(x, "mailuser"):
        walkMailname(visitor, c, parents)

def walkDomain(visitor, x, parents):
    domain = Domain(x)
    visitor.visit(domain, x, parents)

    for c in childElements(x):
        if c.nodeName == "mailsystem":
            walkMailsystem(visitor, c, parents + [domain])
        elif c.nodeName == "database":
            walkDatabase(visitor, c, parents + [domain])
        elif c.nodeName == "maillists":
            walkMaillists(visitor, c, parents + [domain])
        elif c.nodeName == "phosting":
            walkPhosting(visitor, c, parents + [domain])

def walkClient(visitor, x):
    client = Client(x)
    visitor.visit(client, x, [])

    for c in childElementsName(x, "domain"):
        walkDomain(visitor, c, [client])

def walkDump(visitor, fh):
    doc = xml.dom.minidom.parse(fh)
    root = doc.documentElement
    for c in childElements(root):
        if c.nodeName == "client":
            walkClient(visitor, c)
        elif c.nodeName == "domain":
            walkDomain(visitor, c, [])


# Local Variables:
# mode: python
# indent-tabs-mode: nil
# tab-width: 4
# End:
