import sys, os, unpacker, split_files, string, backup

def usage():
    print "Usage: cid-unpacker <session-dir>. Stdin is command."

def getCidFh(sessionDir, cid):
    id = unpacker.Id(cid)
    firstPartName = sessionDir + "/archives/" + id.filename
    if os.path.exists(firstPartName)==0:
	raise Exception, "Could not found file '" + firstPartName + "'!"
    return split_files.SplitFile(backup.BackupPartsFinder("file", firstPartName))

def main():
    if len(sys.argv) != 2:
        usage()
        sys.exit(1)

    sessionDir = sys.argv[1]

    cmd = string.split(sys.stdin.read(), "\0")
    if len(cmd) < 1:
        sys.stderr.write("No command found\n")
        sys.exit(1)

    commandsArgsNum = { "untar": 2,
                        "db" : 7,
                        "unpack" : 2 }

    commands = { "untar" : unpacker.TarUnpacker,
                 "db" : unpacker.DbUnpacker,
                 "unpack" : unpacker.PartUnpacker }

    (command, args) = cmd[0], cmd[1:]

    if command not in commands:
        sys.stderr.write("Unknown command: " + command + "\n")
        sys.exit(1)

    if len(args) != commandsArgsNum[command]:
        sys.stderr.write("Command '" + command + "' takes "
                         + str(commandsArgsNum[command]) + " parameters")
        sys.exit(1)

    (cid, args) = args[0], args[1:]

    #
    # FreeBSD up to 5.3 has bug: read(2) on blocking pipe
    # sometimes return EAGAIN.  GNU tar is not ready to handle
    # this, so we just restart tar four times (arbitrary value, but
    # sounds reasonable) if error occured
    #
    # This workaround should be removed as FreeBSD 4.11 is dropped
    #

    for lastAttempt in (False, False, False, True):
        try:
            fh = getCidFh(sessionDir, cid)

            output = commands[command](*args)(fh, cid)
            sys.stdout.write("\0".join(output))
        except Exception, e:
            sys.stderr.write("Runtime error during " + command + ": " + unicode(e) + "\n")
            if lastAttempt:
                sys.exit(1)
            else:
                sys.stderr.write("One more attempt\n")
        else:
            sys.exit(0)

if __name__ == '__main__':
    main()
