<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
	<xsl:output method="xml" encoding="UTF-8"/>

  <xsl:template match="*" name="copy">
    <xsl:copy>
      <xsl:call-template name="copy-attributes"/>
      <xsl:for-each select="text()">
        <xsl:apply-templates select="."/>
      </xsl:for-each>
      <xsl:for-each select="*">
        <xsl:apply-templates select="."/>
      </xsl:for-each>
    </xsl:copy>
  </xsl:template>

  <xsl:template name="copy-attributes">
    <xsl:for-each select="@*[not(starts-with(name(), 'cid_')) and not(starts-with(name(), 'cid-')) and name()!='cid']">
      <xsl:apply-templates select="."/>
    </xsl:for-each>
    <xsl:if test="count(@*[starts-with(name(), 'cid_') or starts-with(name(), 'cid-') or name()='cid'])>0">
      <content>
        <xsl:for-each select="@*[starts-with(name(), 'cid_') or starts-with(name(), 'cid-')]">
          <cid>
            <xsl:attribute name="type">
              <xsl:choose>
                <xsl:when test="name()='cid_incoming' and name(..)='anonftp'">ftp_incoming</xsl:when>
                <xsl:when test="name()='cid_ftpstat'">ftp_stat</xsl:when>
                <xsl:otherwise>
                  <xsl:value-of select="substring(name(), 5)"/>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:attribute>
            <content-file>
              <xsl:value-of select="."/>
            </content-file>
          </cid>
        </xsl:for-each>
        <xsl:for-each select="@*[name()='cid']">
          <cid>
            <xsl:attribute name="type">
              <xsl:choose>
                <xsl:when test="name(..)='database'">sqldump</xsl:when>
                <xsl:when test="name(..)='key'">key</xsl:when>
                <xsl:when test="name(..)='mailbox'">mailbox</xsl:when>
                <xsl:when test="name(..)='maillist'">mailbox</xsl:when>
                <xsl:when test="name(..)='spamassassin'">spam-assasin</xsl:when>
                <xsl:when test="name(..)='anonftp'">ftp_pub</xsl:when>
                <xsl:when test="name(..)='webuser'">webuser_home</xsl:when>
                <xsl:when test="name(..)='webapp'">web-app</xsl:when>
                <xsl:otherwise>
                  <xsl:message>
                    Unable to determine type of cid at element '<xsl:value-of select="name(..)"/>'
                  </xsl:message>
                  <xsl:value-of select="'unknown'"/>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:attribute>
            <content-file>
              <xsl:value-of select="."/>
            </content-file>
          </cid>
        </xsl:for-each>
      </content>
    </xsl:if>
  </xsl:template>

  <xsl:template match="@*">
    <xsl:copy/>
  </xsl:template>

  <xsl:template match="migration-dump">
    <xsl:copy>
      <xsl:for-each select="@*">
        <xsl:if test="name()!= 'agent-name' and name()!= 'dump-version'">
          <xsl:apply-templates select="."/>
        </xsl:if>
      </xsl:for-each>
      <xsl:attribute name="agent-name">pre9-backup-convert</xsl:attribute>
      <xsl:attribute name="dump-version">9.0</xsl:attribute>

      <xsl:if test="count(client|domain)>0">
        <xsl:choose>

          <xsl:when test="count(client|domain)>1 or count(server)>0">
            <admin>

              <xsl:if test="count(client)>0">
                <clients>
                  <xsl:for-each select="client">
                    <xsl:apply-templates select="."/>
                  </xsl:for-each>
                </clients>
              </xsl:if>

              <xsl:if test="count(domain)>0">
                <domains>
                  <xsl:for-each select="domain">
                    <xsl:apply-templates select="."/>
                  </xsl:for-each>
                </domains>
              </xsl:if>

              <xsl:if test="count(server/admin-password|server/pinfo)>0">
                <preferences>
                  <xsl:for-each select="server/admin-password|server/pinfo">
                    <xsl:apply-templates select="."/>
                  </xsl:for-each>
                </preferences>
              </xsl:if>

            </admin>
          </xsl:when>

          <xsl:otherwise>

            <xsl:for-each select="client">
              <xsl:apply-templates select="."/>
            </xsl:for-each>

            <xsl:for-each select="domain">
              <xsl:apply-templates select="."/>
            </xsl:for-each>

          </xsl:otherwise>

        </xsl:choose>
      </xsl:if>

      <xsl:for-each select="server">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

    </xsl:copy>
  </xsl:template>

  <xsl:template match="server">
    <xsl:copy>
      <xsl:call-template name="copy-attributes"/>

      <xsl:if test="count(system-ip|default-ip|hostname)">
        <properties>
          <xsl:for-each select="system-ip|default-ip|hostname">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </properties>
      </xsl:if>

      <xsl:if test="count(db-server)">
        <db-servers>
          <xsl:for-each select="db-server">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </db-servers>
      </xsl:if>

      <xsl:if test="count(key)>0">
        <keys>
          <xsl:for-each select="key">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </keys>
      </xsl:if>

      <xsl:for-each select="mail-settings">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="dns-settings">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:if test="count(certificate)>0">
        <certificates>
          <xsl:for-each select="certificate">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </certificates>
      </xsl:if>

      <xsl:if test="count(custom-button)>0">
        <interface-preferences>
          <xsl:for-each select="custom-button">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </interface-preferences>
      </xsl:if>

      <xsl:if test="count(client-template|domain-template)>0">
        <account-templates>
          <xsl:for-each select="client-template|domain-template">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </account-templates>
      </xsl:if>

      <xsl:for-each select="sso-settings">
        <sso-settings>
          <xsl:if test="@enabled">
            <xsl:copy-of select="@enabled"/>
          </xsl:if>
          <xsl:apply-templates select="./*"/>
        </sso-settings>
      </xsl:for-each>

      <xsl:for-each select="application-vault">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="sb-config">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="server-preferences">
        <xsl:apply-templates select="."/>
        <xsl:for-each select="../force-db-prefix">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </xsl:for-each>

    </xsl:copy>
  </xsl:template>
  
  <xsl:template match="sapp-package">
    <xsl:copy>
      <xsl:call-template name="copy-attributes"/>

      <xsl:if test="sapp-distrib">
        <content>
          <cid type="sapp-distrib">
            <content-file>
              <xsl:value-of select="sapp-distrib/text()"/>
            </content-file>
          </cid>
        </content>
      </xsl:if>

      <sapp-name>
        <xsl:value-of select="sapp-name/text()"/>
      </sapp-name>

      <xsl:if test="sapp-version">
        <xsl:copy-of select="sapp-version"/>
      </xsl:if>

      <xsl:if test="sapp-release">
        <xsl:copy-of select="sapp-release"/>
      </xsl:if>

    </xsl:copy>
  </xsl:template>
  
  <xsl:template match="client">
    <xsl:copy>
      <xsl:call-template name="copy-attributes"/>
      
      <preferences>
        <xsl:for-each select="custom-button|domain-template|pinfo">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </preferences>

      <properties>
        <xsl:for-each select="password|status|sb-reseller">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </properties>

      <xsl:if test="count(limit|permission)>0">
        <limits-and-permissions>
          <xsl:for-each select="limit|permission">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </limits-and-permissions>
      </xsl:if>

      <xsl:for-each select="ip_pool">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="sapp-pool">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:if test="count(domain)>0">
        <domains>
          <xsl:for-each select="domain">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </domains>
      </xsl:if>

    </xsl:copy>
  </xsl:template>

  <xsl:template match="domain">
    <xsl:copy>
      <xsl:call-template name="copy-attributes"/>

      <xsl:if test="../@guid">
        <xsl:attribute name="owner-guid">
          <xsl:value-of select="../@guid"/>
        </xsl:attribute>
      </xsl:if>

      <preferences>
        <xsl:for-each select="default-domain">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
        <xsl:for-each select="domain-alias">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
        <xsl:for-each select="custom-button">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </preferences>

      <properties>
        <xsl:for-each select="ip|status|dns-zone">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </properties>

      <xsl:if test="count(limit)>0">
        <limits-and-permissions>
          <xsl:for-each select="limit">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </limits-and-permissions>
      </xsl:if>

      <xsl:for-each select="mailsystem">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:if test="count(database)>0">
        <databases>
          <xsl:for-each select="database">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </databases>
      </xsl:if>

      <xsl:for-each select="maillists">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="traffic">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:if test="count(certificate)>0">
        <certificates>
          <xsl:for-each select="certificate">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </certificates>
      </xsl:if>

      <xsl:for-each select="tomcat">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="domainuser">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="phosting|shosting|fhosting">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

    </xsl:copy>
  </xsl:template>

  <xsl:template match="mailsystem">
    <xsl:copy>
      
      <properties>
        <xsl:for-each select="status">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </properties>

      <xsl:if test="count(mailuser)>0">
        <mailusers>
          <xsl:for-each select="mailuser">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </mailusers>
      </xsl:if>

      <xsl:if test="count(catch-all|domain-keys)>0">
        <preferences>
          <xsl:for-each select="catch-all|domain-keys">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </preferences>
      </xsl:if>

    </xsl:copy>
  </xsl:template>

  <xsl:template match="phosting">
    <xsl:copy>

      <xsl:if test="../@guid">
        <xsl:attribute name="owner-guid">
          <xsl:value-of select="../@guid"/>
        </xsl:attribute>
      </xsl:if>

      <xsl:call-template name="copy-attributes"/>

      <preferences>
        <xsl:for-each select="sysuser|fpuser|logrotation|anonftp|pdir|webalizer|sb-domain|perfomance">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </preferences>

      <xsl:if test="count(scripting)>0">
        <limits-and-permissions>
          <xsl:for-each select="scripting">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </limits-and-permissions>
      </xsl:if>

      <xsl:if test="count(sapp-installed)>0">
        <applications>
          <xsl:for-each select="sapp-installed">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </applications>
      </xsl:if>

      <xsl:if test="count(webuser)>0">
        <webusers>
          <xsl:for-each select="webuser">
			<xsl:element name="webuser">
				<xsl:attribute name="name">
					<xsl:value-of select="./sysuser/@name"/>
				</xsl:attribute>
				<xsl:call-template name="copy-attributes"/>
				<xsl:for-each select="sysuser|scripting">
					<xsl:apply-templates select="."/>
				</xsl:for-each>
			</xsl:element>
          </xsl:for-each>
        </webusers>
      </xsl:if>

      <xsl:if test="count(subdomain)>0">
        <subdomains>
          <xsl:for-each select="subdomain">
            <xsl:apply-templates select="."/>
          </xsl:for-each>
        </subdomains>
      </xsl:if>

    </xsl:copy>
  </xsl:template>

  <xsl:template match="subdomain">
    <xsl:if test="../@guid">
      <xsl:attribute name="owner-guid">
        <xsl:value-of select="../@guid"/>
      </xsl:attribute>
    </xsl:if>

    <xsl:call-template name="copy"/>
  </xsl:template>

  <xsl:template match="mailuser">
    <mailuser>

      <xsl:if test="../../@guid">
        <xsl:attribute name="owner-guid">
          <xsl:value-of select="../../@guid"/>
        </xsl:attribute>
      </xsl:if>

      <xsl:call-template name="copy-attributes"/>

      <properties>
        <xsl:for-each select="password">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </properties>

      <limits-and-permissions>
        <xsl:for-each select="mailuser-permission">
          <xsl:apply-templates select="."/>
        </xsl:for-each>
      </limits-and-permissions>

      <preferences>
        <xsl:for-each select="*[name()!='password' and name()!='mailuser-permission']">
			<xsl:choose>
				<xsl:when test="name() = 'drweb'">
					<virusfilter><xsl:call-template name="copy-attributes"/></virusfilter>
				</xsl:when>
				<xsl:otherwise>
					<xsl:apply-templates select="."/>
				</xsl:otherwise>
			</xsl:choose>
        </xsl:for-each>
      </preferences>
      </mailuser>
    </xsl:template>

  <xsl:template match="database">
    <xsl:copy>

      <xsl:if test="../@guid">
        <xsl:attribute name="owner-guid">
          <xsl:value-of select="../@guid"/>
        </xsl:attribute>
      </xsl:if>

      <xsl:call-template name="copy-attributes"/>

      <xsl:for-each select="text()">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

      <xsl:for-each select="*">
        <xsl:apply-templates select="."/>
      </xsl:for-each>

    </xsl:copy>
  </xsl:template>

  <xsl:template match="permission|mailuser-permission">
    <xsl:element name="{name()}">
      <xsl:attribute name="name">
		  <xsl:choose>
			<xsl:when test="@name='manage_drweb'">manage_virusfilter</xsl:when>
			<xsl:otherwise>
				<xsl:value-of select="@name"/>
			</xsl:otherwise>
		  </xsl:choose>
      </xsl:attribute>
      <xsl:attribute name="value">
        <xsl:value-of select="@allowed"/>
      </xsl:attribute>
    </xsl:element>
  </xsl:template>
  
  <xsl:template match="admin">
    <db-admin>
      <xsl:call-template name="copy-attributes"/>
      <xsl:for-each select="*">
        <xsl:apply-templates select="."/>
      </xsl:for-each>
    </db-admin>
  </xsl:template>

  <xsl:template match="spamassassin/@action">
    <xsl:attribute name="action">
      <xsl:choose>
        <xsl:when test=".='del'">delete</xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="."/>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:attribute>
  </xsl:template>

  <xsl:template match="sapp-installed">
    <sapp-installed>
      <xsl:if test="count(sapp-scripts|sapp-https-scripts|sapp-cgi-bin|sapp-htdocs)>0">
        <content>
          <xsl:for-each select="sapp-scripts|sapp-https-scripts|sapp-cgi-bin|sapp-htdocs">
            <cid>
              <xsl:attribute name="type">
                <xsl:value-of select="name()"/>
              </xsl:attribute>
              <content-file>
                <xsl:value-of select="."/>
              </content-file>
            </cid>
          </xsl:for-each>
        </content>
      </xsl:if>
      <xsl:apply-templates select="sapp-spec"/>
      <xsl:apply-templates select="sapp-param"/>
      <xsl:apply-templates select="database"/>
      <xsl:apply-templates select="custom-button"/>
      <xsl:apply-templates select="sapp-installdir"/>
    </sapp-installed>
  </xsl:template>
  
</xsl:stylesheet>
