import os, re


SPEC_FILE = "/control/spec"
ENTRIES = [ "Description", "EntryPoint", "DocumentType", "Stylesheet", "Interpreter" ]


class ParseFailed(Exception):
    def __init__(self, spec, lineNumber, line):
        self.spec = spec
        self.lineNumber = lineNumber
        self.line = line
    def __str__(self):
        return "%s:%s:Expected '<token>: <value>', got '%s'" \
               % (self.spec, self.lineNumber, self.line)


class IllegalParameter(Exception):
    def __init__(self, spec, key):
        self.spec = spec
        self.key = key
    def __str__(self):
        return "Illegal parameter found in %s:%s" % (self.spec, self.key)


class AgentSpec(object):
    def __init__(self, agentDir):
        self.__parameters = self.__parse(agentDir + SPEC_FILE)

    def __len__(self): return len(self.__parameters)
    def __getitem__(self, key): return self.__parameters[key]
    def __iter__(self): return self.iterkeys()
    def iterkeys(self): return self.__parameters.iterkeys()
    def iteritems(self): return self.__parameters.iteritems()

    def __parse(self, specFile):
        parameters = {}

        spec = file(specFile)
        lineNumber = 0
        paramRe = re.compile("^([^:]+?):\\s(.*)\\n$")
        for line in spec:
            lineNumber += 1
            if line == "":
                continue
            res = paramRe.match(line)
            if not res:
                raise ParseFailed(specFile, lineNumber, line)
            (key, value) = res.groups()
            if key not in ENTRIES:
                raise IllegalParameter(specFile, key)
            parameters[key] = value
        return parameters


def parseAgents(agentsDir):
    agents = []
    for f in os.listdir(agentsDir):
        d = os.path.join(agentsDir, f)
        if os.path.isdir(d):
            try:
                agents += [AgentSpec(d)]
            except:
                pass
    return agents


