import stacktrace
import re

REQUEST_TEXT=\
"""==HTTP request==========================
%(method)s %(url)s
%(headers)s

%(body)s
========================================
"""

def __headersToStr(headers):
    return "\n".join([n + ": " + v for n, v in headers.items()])

def __getRequestText(request):
    if request.headers.has_key('Content-type') \
            and re.match('^multipart/form-data', request.headers['Content-type']):
        body = "<request body>"
    else:
        body = request.get_data()

    return REQUEST_TEXT % {
        "url": request.get_full_url(),
        "method":  request.get_method(),
        "headers": __headersToStr(request.headers),
        "body": body
    }


HTTP_ERROR_HEADER="""HTTP error during API-RPC call.
"""
HTTP_ERROR_DESCRIPTION=\
"""==HTTP response=========================
%(code)d %(status)s
%(headers)s

%(body)s
========================================
"""

URL_ERROR_HEADER="""Socket error during API-RPC call.
"""
URL_ERROR_DESCRIPTION=\
"""Socket error: %(error)s.
"""

API_RPC_PARSE_ERROR_HEADER="""Unable to parse API-RPC response.
"""
API_RPC_PARSE_ERROR_DESCRIPTION=\
"""Error during parsing API-RPC response:
%(errormsg)s
==HTTP response body====================
%(body)s
========================================
"""

API_RPC_ERROR_HEADER="""API-RPC call finished with an error.
"""
API_RPC_ERROR_DESCRIPTION=\
"""Error code: %(code)s.
Error description:
%(desc)s
"""

GENERIC_ERROR_HEADER="""Internal error during migration.
"""
GENERIC_ERROR_DESCRIPTION=\
"""Stacktrace:
%(stacktrace)s
"""

def __getMessageHeader(type, **kvargs):
    return globals()[type + "_ERROR_HEADER"] % kvargs

def __getMessageDesc(type, **kvargs):
    return globals()[type + "_ERROR_DESCRIPTION"] % kvargs

def __getErrorMessage(type, request, **kvargs):
    return __getMessageHeader(type, **kvargs) \
           + __getRequestText(request) \
           + __getMessageDesc(type, **kvargs)

def genericErrorMessage(req):
    return __getErrorMessage("GENERIC", req, stacktrace = stacktrace.stacktrace())

def urlErrorMessage(req, urlError):
    return __getErrorMessage("URL", req, error = urlError.reason)

def httpErrorMessage(req, httpError):
    if httpError.code == -1:
        return __getErrorMessage("URL", req, error = "Remote server closed connection")
    else:
        return __getErrorMessage("HTTP", req, code = httpError.code,
                                 status = httpError.msg,
                                 headers = str(httpError.info()),
                                 body = httpError.read())

def parsingErrorMessage(req, parseError, responseContent):
    return __getErrorMessage("API_RPC_PARSE", req,
                             errormsg = "\n".join([i for i in parseError]),
                             body = responseContent)

def apiRpcErrorMessage(req, answer):
    return __getErrorMessage("API_RPC", req, code = answer.getErrcode(), desc = answer.getErrtext())

